#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_security_group
version_added: 1.0.0
author:
  - "Andrew de Quincey (@adq)"
  - "Razique Mahroua (@Razique)"
short_description: Maintain an EC2 security group
description:
  - Maintains EC2 security groups.
options:
  name:
    description:
      - Name of the security group.
      - One of and only one of I(name) or I(group_id) is required.
      - Required if I(state=present).
    required: false
    type: str
  group_id:
    description:
      - Id of group to delete (works only with absent).
      - One of and only one of I(name) or I(group_id) is required.
    required: false
    type: str
  description:
    description:
      - Description of the security group.
      - Required when I(state) is C(present).
    required: false
    type: str
  vpc_id:
    description:
      - ID of the VPC to create the group in.
    required: false
    type: str
  rules:
    description:
      - List of firewall inbound rules to enforce in this group (see example). If none are supplied,
        no inbound rules will be enabled. Rules list may include its own name in I(group_name).
        This allows idempotent loopback additions (e.g. allow group to access itself).
    required: false
    type: list
    elements: dict
    suboptions:
        cidr_ip:
          type: list
          elements: raw
          description:
            - The IPv4 CIDR range traffic is coming from.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
            - Support for passing nested lists of strings to I(cidr_ip) has been deprecated and will
              be removed in a release after 2024-12-01.
        cidr_ipv6:
          type: list
          elements: raw
          description:
            - The IPv6 CIDR range traffic is coming from.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
            - Support for passing nested lists of strings to I(cidr_ipv6) has been deprecated and will
              be removed in a release after 2024-12-01.
        ip_prefix:
          type: list
          elements: str
          description:
            - The IP Prefix U(https://docs.aws.amazon.com/cli/latest/reference/ec2/describe-prefix-lists.html)
              that traffic is coming from.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
        group_id:
          type: list
          elements: str
          description:
            - The ID of the Security Group that traffic is coming from.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
        group_name:
          type: list
          elements: str
          description:
            - Name of the Security Group that traffic is coming from.
            - If the Security Group doesn't exist a new Security Group will be
              created with I(group_desc) as the description.
            - I(group_name) can accept values of type str and list.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
        group_desc:
          type: str
          description:
            - If the I(group_name) is set and the Security Group doesn't exist a new Security Group will be
              created with I(group_desc) as the description.
        proto:
          type: str
          description:
            - The IP protocol name (C(tcp), C(udp), C(icmp), C(icmpv6)) or
              number (U(https://en.wikipedia.org/wiki/List_of_IP_protocol_numbers)).
          default: 'tcp'
        from_port:
          type: int
          description:
            - The start of the range of ports that traffic is going to.
            - A value can be between C(0) to C(65535).
            - When I(proto=icmp) a value of C(-1) indicates all ports.
            - Mutually exclusive with I(icmp_code), I(icmp_type) and I(ports).
        to_port:
          type: int
          description:
            - The end of the range of ports that traffic is going to.
            - A value can be between C(0) to C(65535).
            - When I(proto=icmp) a value of C(-1) indicates all ports.
            - Mutually exclusive with I(icmp_code), I(icmp_type) and I(ports).
        ports:
          type: list
          elements: str
          description:
            - A list of ports that traffic is going to.
            - Elements of the list can be a single port (for example C(8080)), or a range of ports
              specified as C(<START>-<END>), (for example C(1011-1023)).
            - Mutually exclusive with I(icmp_code), I(icmp_type), I(from_port) and I(to_port).
        icmp_type:
          version_added: 3.3.0
          type: int
          description:
            - The ICMP type of the packet.
            - A value of C(-1) indicates all ICMP types.
            - Requires I(proto=icmp) or I(proto=icmpv6).
            - Mutually exclusive with I(ports), I(from_port) and I(to_port).
        icmp_code:
          version_added: 3.3.0
          type: int
          description:
            - The ICMP code of the packet.
            - A value of C(-1) indicates all ICMP codes.
            - Requires I(proto=icmp) or I(proto=icmpv6).
            - Mutually exclusive with I(ports), I(from_port) and I(to_port).
        rule_desc:
          type: str
          description: A description for the rule.

  rules_egress:
    description:
      - List of firewall outbound rules to enforce in this group (see example). If none are supplied,
        a default all-out rule is assumed. If an empty list is supplied, no outbound rules will be enabled.
    required: false
    type: list
    elements: dict
    aliases: ['egress_rules']
    suboptions:
        cidr_ip:
          type: list
          elements: raw
          description:
            - The IPv4 CIDR range traffic is going to.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
            - Support for passing nested lists of strings to I(cidr_ip) has been deprecated and will
              be removed in a release after 2024-12-01.
        cidr_ipv6:
          type: list
          elements: raw
          description:
            - The IPv6 CIDR range traffic is going to.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
            - Support for passing nested lists of strings to I(cidr_ipv6) has been deprecated and will
              be removed in a release after 2024-12-01.
        ip_prefix:
          type: list
          elements: str
          description:
            - The IP Prefix U(https://docs.aws.amazon.com/cli/latest/reference/ec2/describe-prefix-lists.html)
              that traffic is going to.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
        group_id:
          type: list
          elements: str
          description:
            - The ID of the Security Group that traffic is going to.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
        group_name:
          type: list
          elements: str
          description:
            - Name of the Security Group that traffic is going to.
            - If the Security Group doesn't exist a new Security Group will be
              created with I(group_desc) as the description.
            - You can specify only one of I(cidr_ip), I(cidr_ipv6), I(ip_prefix), I(group_id)
              and I(group_name).
        group_desc:
          type: str
          description:
            - If the I(group_name) is set and the Security Group doesn't exist a new Security Group will be
              created with I(group_desc) as the description.
        proto:
          type: str
          description:
            - The IP protocol name (C(tcp), C(udp), C(icmp), C(icmpv6)) or
              number (U(https://en.wikipedia.org/wiki/List_of_IP_protocol_numbers)).
          default: 'tcp'
        from_port:
          type: int
          description:
            - The start of the range of ports that traffic is going to.
            - A value can be between C(0) to C(65535).
            - When I(proto=icmp) a value of C(-1) indicates all ports.
            - Mutually exclusive with I(icmp_code), I(icmp_type) and I(ports).
        to_port:
          type: int
          description:
            - The end of the range of ports that traffic is going to.
            - A value can be between C(0) to C(65535).
            - When I(proto=icmp) a value of C(-1) indicates all ports.
            - Mutually exclusive with I(icmp_code), I(icmp_type) and I(ports).
        ports:
          type: list
          elements: str
          description:
            - A list of ports that traffic is going to.
            - Elements of the list can be a single port (for example C(8080)), or a range of ports
              specified as C(<START>-<END>), (for example C(1011-1023)).
            - Mutually exclusive with I(icmp_code), I(icmp_type), I(from_port) and I(to_port).
        icmp_type:
          version_added: 3.3.0
          type: int
          description:
            - The ICMP type of the packet.
            - A value of C(-1) indicates all ICMP types.
            - Requires I(proto=icmp) or I(proto=icmpv6).
            - Mutually exclusive with I(ports), I(from_port) and I(to_port).
        icmp_code:
          version_added: 3.3.0
          type: int
          description:
            - The ICMP code of the packet.
            - A value of C(-1) indicates all ICMP codes.
            - Requires I(proto=icmp) or I(proto=icmpv6).
            - Mutually exclusive with I(ports), I(from_port) and I(to_port).
        rule_desc:
            type: str
            description: A description for the rule.
  state:
    description:
      - Create or delete a security group.
    required: false
    default: 'present'
    choices: [ "present", "absent" ]
    aliases: []
    type: str
  purge_rules:
    description:
      - Purge existing rules on security group that are not found in rules.
    required: false
    default: 'true'
    aliases: []
    type: bool
  purge_rules_egress:
    description:
      - Purge existing rules_egress on security group that are not found in rules_egress.
    required: false
    default: 'true'
    aliases: ['purge_egress_rules']
    type: bool

extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3

notes:
  - If a rule declares a group_name and that group doesn't exist, it will be
    automatically created. In that case, group_desc should be provided as well.
    The module will refuse to create a depended-on group without a description.
  - Prior to release 5.0.0 this module was called C(amazon.aws.ec2_group_info).  The usage did not
    change.
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: example using security group rule descriptions
  amazon.aws.ec2_security_group:
    name: "{{ name }}"
    description: sg with rule descriptions
    vpc_id: vpc-xxxxxxxx
    rules:
      - proto: tcp
        ports:
          - 80
        cidr_ip: 0.0.0.0/0
        rule_desc: allow all on port 80

- name: example using ICMP types and codes
  amazon.aws.ec2_security_group:
    name: "{{ name }}"
    description: sg for ICMP
    vpc_id: vpc-xxxxxxxx
    rules:
      - proto: icmp
        icmp_type: 3
        icmp_code: 1
        cidr_ip: 0.0.0.0/0

- name: example ec2 group
  amazon.aws.ec2_security_group:
    name: example
    description: an example EC2 group
    vpc_id: 12345
    rules:
      - proto: tcp
        from_port: 80
        to_port: 80
        cidr_ip: 0.0.0.0/0
      - proto: tcp
        from_port: 22
        to_port: 22
        cidr_ip: 10.0.0.0/8
      - proto: tcp
        from_port: 443
        to_port: 443
        # this should only be needed for EC2 Classic security group rules
        # because in a VPC an ELB will use a user-account security group
        group_id: amazon-elb/sg-87654321/amazon-elb-sg
      - proto: tcp
        from_port: 3306
        to_port: 3306
        group_id: 123456789012/sg-87654321/exact-name-of-sg
      - proto: udp
        from_port: 10050
        to_port: 10050
        cidr_ip: 10.0.0.0/8
      - proto: udp
        from_port: 10051
        to_port: 10051
        group_id: sg-12345678
      - proto: icmp
        from_port: 8 # icmp type, -1 = any type
        to_port: -1 # icmp subtype, -1 = any subtype
        cidr_ip: 10.0.0.0/8
      - proto: all
        # the containing group name may be specified here
        group_name: example
      - proto: all
        # in the 'proto' attribute, if you specify -1 (only supported when I(proto=icmp)), all, or a protocol number
        # other than tcp, udp, icmp, or 58 (ICMPv6), traffic on all ports is allowed, regardless of any ports that
        # you specify.
        from_port: 10050 # this value is ignored
        to_port: 10050   # this value is ignored
        cidr_ip: 10.0.0.0/8

    rules_egress:
      - proto: tcp
        from_port: 80
        to_port: 80
        cidr_ip: 0.0.0.0/0
        cidr_ipv6: 64:ff9b::/96
        group_name: example-other
        # description to use if example-other needs to be created
        group_desc: other example EC2 group

- name: example2 ec2 group
  amazon.aws.ec2_security_group:
    name: example2
    description: an example2 EC2 group
    vpc_id: 12345
    rules:
      # 'ports' rule keyword was introduced in version 2.4. It accepts a single
      # port value or a list of values including ranges (from_port-to_port).
      - proto: tcp
        ports: 22
        group_name: example-vpn
      - proto: tcp
        ports:
          - 80
          - 443
          - 8080-8099
        cidr_ip: 0.0.0.0/0
      # Rule sources list support was added in version 2.4. This allows to
      # define multiple sources per source type as well as multiple source types per rule.
      - proto: tcp
        ports:
          - 6379
          - 26379
        group_name:
          - example-vpn
          - example-redis
      - proto: tcp
        ports: 5665
        group_name: example-vpn
        cidr_ip:
          - 172.16.1.0/24
          - 172.16.17.0/24
        cidr_ipv6:
          - 2607:F8B0::/32
          - 64:ff9b::/96
        group_id:
          - sg-edcd9784
  diff: true

- name: "Delete group by its id"
  amazon.aws.ec2_security_group:
    group_id: sg-33b4ee5b
    state: absent
"""

RETURN = r"""
group_name:
  description: Security group name
  sample: My Security Group
  type: str
  returned: on create/update
group_id:
  description: Security group id
  sample: sg-abcd1234
  type: str
  returned: on create/update
description:
  description: Description of security group
  sample: My Security Group
  type: str
  returned: on create/update
tags:
  description: Tags associated with the security group
  sample:
    Name: My Security Group
    Purpose: protecting stuff
  type: dict
  returned: on create/update
vpc_id:
  description: ID of VPC to which the security group belongs
  sample: vpc-abcd1234
  type: str
  returned: on create/update
ip_permissions:
  description: Inbound rules associated with the security group.
  sample:
    - from_port: 8182
      ip_protocol: tcp
      ip_ranges:
        - cidr_ip: "198.51.100.1/32"
      ipv6_ranges: []
      prefix_list_ids: []
      to_port: 8182
      user_id_group_pairs: []
  type: list
  returned: on create/update
ip_permissions_egress:
  description: Outbound rules associated with the security group.
  sample:
    - ip_protocol: -1
      ip_ranges:
        - cidr_ip: "0.0.0.0/0"
          ipv6_ranges: []
          prefix_list_ids: []
          user_id_group_pairs: []
  type: list
  returned: on create/update
owner_id:
  description: AWS Account ID of the security group
  sample: 123456789012
  type: int
  returned: on create/update
"""

import itertools
import json
import re
from collections import namedtuple
from copy import deepcopy
from ipaddress import ip_network
from time import sleep

try:
    import botocore
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils._text import to_text
from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict
from ansible.module_utils.common.network import to_ipv6_subnet
from ansible.module_utils.common.network import to_subnet
from ansible.module_utils.six import string_types

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.iam import get_aws_account_id
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.tagging import compare_aws_tags
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list
from ansible_collections.amazon.aws.plugins.module_utils.transformation import scrub_none_parameters
from ansible_collections.amazon.aws.plugins.module_utils.waiters import get_waiter

Rule = namedtuple("Rule", ["port_range", "protocol", "target", "target_type", "description"])
TARGET_TYPES_ALL = {"ipv4", "ipv6", "group", "ip_prefix"}
SOURCE_TYPES_ALL = {"cidr_ip", "cidr_ipv6", "group_id", "group_name", "ip_prefix"}
PORT_TYPES_ALL = {"from_port", "to_port", "ports", "icmp_type", "icmp_code"}
current_account_id = None


class SecurityGroupError(Exception):
    def __init__(self, msg, e=None, **kwargs):
        super().__init__(msg)
        self.message = msg
        self.exception = e
        self.kwargs = kwargs

    # Simple helper to perform the module.fail_... call once we have module available to us
    def fail(self, module):
        if self.exception:
            module.fail_json_aws(self.exception, msg=self.message, **self.kwargs)
        module.fail_json(msg=self.message, **self.kwargs)


def rule_cmp(a, b):
    """Compare rules without descriptions"""
    for prop in ["port_range", "protocol", "target", "target_type"]:
        if prop == "port_range" and to_text(a.protocol) == to_text(b.protocol):
            # equal protocols can interchange `(-1, -1)` and `(None, None)`
            if a.port_range in ((None, None), (-1, -1)) and b.port_range in ((None, None), (-1, -1)):
                continue
            elif getattr(a, prop) != getattr(b, prop):
                return False
        elif getattr(a, prop) != getattr(b, prop):
            return False
    return True


def rules_to_permissions(rules):
    return [to_permission(rule) for rule in rules]


def to_permission(rule):
    # take a Rule, output the serialized grant
    perm = {
        "IpProtocol": rule.protocol,
    }
    perm["FromPort"], perm["ToPort"] = rule.port_range
    if rule.target_type == "ipv4":
        perm["IpRanges"] = [
            {
                "CidrIp": rule.target,
            }
        ]
        if rule.description:
            perm["IpRanges"][0]["Description"] = rule.description
    elif rule.target_type == "ipv6":
        perm["Ipv6Ranges"] = [
            {
                "CidrIpv6": rule.target,
            }
        ]
        if rule.description:
            perm["Ipv6Ranges"][0]["Description"] = rule.description
    elif rule.target_type == "group":
        if isinstance(rule.target, tuple):
            pair = {}
            if rule.target[0]:
                pair["UserId"] = rule.target[0]
            # group_id/group_name are mutually exclusive - give group_id more precedence as it is more specific
            if rule.target[1]:
                pair["GroupId"] = rule.target[1]
            elif rule.target[2]:
                pair["GroupName"] = rule.target[2]
            perm["UserIdGroupPairs"] = [pair]
        else:
            perm["UserIdGroupPairs"] = [{"GroupId": rule.target}]
        if rule.description:
            perm["UserIdGroupPairs"][0]["Description"] = rule.description
    elif rule.target_type == "ip_prefix":
        perm["PrefixListIds"] = [
            {
                "PrefixListId": rule.target,
            }
        ]
        if rule.description:
            perm["PrefixListIds"][0]["Description"] = rule.description
    elif rule.target_type not in TARGET_TYPES_ALL:
        raise ValueError(f"Invalid target type for rule {rule}")
    return fix_port_and_protocol(perm)


def rule_from_group_permission(perm):
    """
    Returns a rule dict from an existing security group.

    When using a security group as a target all 3 fields (OwnerId, GroupId, and
    GroupName) need to exist in the target. This ensures consistency of the
    values that will be compared to desired_ingress or desired_egress
    in wait_for_rule_propagation().
    GroupId is preferred as it is more specific except when targeting 'amazon-'
    prefixed security groups (such as EC2 Classic ELBs).
    """

    def ports_from_permission(p):
        if "FromPort" not in p and "ToPort" not in p:
            return (None, None)
        return (int(perm["FromPort"]), int(perm["ToPort"]))

    # outputs a rule tuple
    for target_key, target_subkey, target_type in [
        ("IpRanges", "CidrIp", "ipv4"),
        ("Ipv6Ranges", "CidrIpv6", "ipv6"),
        ("PrefixListIds", "PrefixListId", "ip_prefix"),
    ]:
        if target_key not in perm:
            continue
        for r in perm[target_key]:
            # there may be several IP ranges here, which is ok
            yield Rule(
                ports_from_permission(perm),
                to_text(perm["IpProtocol"]),
                r[target_subkey],
                target_type,
                r.get("Description"),
            )
    if "UserIdGroupPairs" in perm and perm["UserIdGroupPairs"]:
        for pair in perm["UserIdGroupPairs"]:
            target = (
                pair.get("UserId", current_account_id),
                pair.get("GroupId", None),
                None,
            )
            if pair.get("UserId", "").startswith("amazon-"):
                # amazon-elb and amazon-prefix rules don't need
                # group-id specified, so remove it when querying
                # from permission
                target = (
                    pair.get("UserId", None),
                    None,
                    pair.get("GroupName", None),
                )
            elif "VpcPeeringConnectionId" not in pair and pair["UserId"] != current_account_id:
                # EC2-Classic cross-account
                pass
            elif "VpcPeeringConnectionId" in pair:
                # EC2-VPC cross-account VPC peering
                target = (
                    pair.get("UserId", None),
                    pair.get("GroupId", None),
                    None,
                )

            yield Rule(
                ports_from_permission(perm), to_text(perm["IpProtocol"]), target, "group", pair.get("Description")
            )


# Wrap just this method so we can retry on missing groups
@AWSRetry.jittered_backoff(retries=5, delay=5, catch_extra_error_codes=["InvalidGroup.NotFound"])
def get_security_groups_with_backoff(client, **kwargs):
    return client.describe_security_groups(**kwargs)


def sg_exists_with_backoff(client, **kwargs):
    try:
        return client.describe_security_groups(aws_retry=True, **kwargs)
    except is_boto3_error_code("InvalidGroup.NotFound"):
        return {"SecurityGroups": []}


def deduplicate_rules_args(rules):
    """Returns unique rules"""
    if rules is None:
        return None
    return list(dict(zip((json.dumps(r, sort_keys=True) for r in rules), rules)).values())


def validate_rule(rule):
    icmp_type = rule.get("icmp_type", None)
    icmp_code = rule.get("icmp_code", None)
    proto = rule["proto"]
    if (icmp_type is not None or icmp_code is not None) and ("icmp" not in proto):
        raise SecurityGroupError(msg="Specify proto: icmp or icmpv6 when using icmp_type/icmp_code")


def _target_from_rule_with_group_id(rule, groups):
    owner_id = current_account_id
    FOREIGN_SECURITY_GROUP_REGEX = r"^([^/]+)/?(sg-\S+)?/(\S+)"
    foreign_rule = re.match(FOREIGN_SECURITY_GROUP_REGEX, rule["group_id"])

    if not foreign_rule:
        return "group", (owner_id, rule["group_id"], None), False

    # this is a foreign Security Group. Since you can't fetch it you must create an instance of it
    # Matches on groups like amazon-elb/sg-5a9c116a/amazon-elb-sg, amazon-elb/amazon-elb-sg,
    # and peer-VPC groups like 0987654321/sg-1234567890/example
    owner_id, group_id, group_name = foreign_rule.groups()
    group_instance = dict(UserId=owner_id, GroupId=group_id, GroupName=group_name)
    groups[group_id] = group_instance
    groups[group_name] = group_instance
    if group_id and group_name:
        if group_name.startswith("amazon-"):
            # amazon-elb and amazon-prefix rules don't need group_id specified,
            group_id = None
        else:
            # For cross-VPC references we'll use group_id as it is more specific
            group_name = None
    return "group", (owner_id, group_id, group_name), False


def _lookup_target_or_fail(client, group_name, vpc_id, groups, msg):
    owner_id = current_account_id
    filters = {"group-name": group_name}
    if vpc_id:
        filters["vpc-id"] = vpc_id

    filters = ansible_dict_to_boto3_filter_list(filters)
    try:
        found_group = get_security_groups_with_backoff(client, Filters=filters).get("SecurityGroups", [])[0]
    except (is_boto3_error_code("InvalidGroup.NotFound"), IndexError):
        raise SecurityGroupError(msg=msg)
    except (BotoCoreError, ClientError) as e:  # pylint: disable=duplicate-except
        raise SecurityGroupError(msg="Failed to get security group", e=e)

    group_id = found_group["GroupId"]
    groups[group_id] = found_group
    groups[group_name] = found_group
    return "group", (owner_id, group_id, None), False


def _create_target_from_rule(client, rule, groups, vpc_id, tags, check_mode):
    owner_id = current_account_id
    # We can't create a group in check mode...
    if check_mode:
        return "group", (owner_id, None, None), True

    group_name = rule["group_name"]

    try:
        created_group = _create_security_group_with_wait(client, group_name, rule["group_desc"], vpc_id, tags)
    except is_boto3_error_code("InvalidGroup.Duplicate"):
        # The group exists, but didn't show up in any of our previous describe-security-groups calls
        # Try searching on a filter for the name, and allow a retry window for AWS to update
        # the model on their end.
        fail_msg = (
            f"Could not create or use existing group '{group_name}' in rule {rule}.  "
            "Make sure the group exists and try using the group_id "
            "instead of the name"
        )
        return _lookup_target_or_fail(client, group_name, vpc_id, groups, fail_msg)
    except (BotoCoreError, ClientError) as e:
        raise SecurityGroupError(msg="Failed to create security group '{0}' in rule {1}", e=e)

    group_id = created_group["GroupId"]
    groups[group_id] = created_group
    groups[group_name] = created_group

    return "group", (owner_id, group_id, None), True


def _target_from_rule_with_group_name(client, rule, name, group, groups, vpc_id, tags, check_mode):
    group_name = rule["group_name"]
    owner_id = current_account_id
    if group_name == name:
        # Simplest case, the rule references itself
        group_id = group["GroupId"]
        groups[group_id] = group
        groups[group_name] = group
        return "group", (owner_id, group_id, None), False

    # Already cached groups
    if group_name in groups and group.get("VpcId") and groups[group_name].get("VpcId"):
        # both are VPC groups, this is ok
        group_id = groups[group_name]["GroupId"]
        return "group", (owner_id, group_id, None), False

    if group_name in groups and not (group.get("VpcId") or groups[group_name].get("VpcId")):
        # both are EC2 classic, this is ok
        group_id = groups[group_name]["GroupId"]
        return "group", (owner_id, group_id, None), False

    # if we got here, either the target group does not exist, or there
    # is a mix of EC2 classic + VPC groups. Mixing of EC2 classic + VPC
    # is bad, so we have to create a new SG because no compatible group
    # exists

    # Without a group description we can't create a new group, try looking up the group, or fail
    # with a descriptive error message
    if not rule.get("group_desc", "").strip():
        # retry describing the group
        fail_msg = (
            f"group '{group_name}' not found and would be automatically created by rule {rule} but "
            "no description was provided"
        )
        return _lookup_target_or_fail(client, group_name, vpc_id, groups, fail_msg)

    return _create_target_from_rule(client, rule, groups, vpc_id, tags, check_mode)


def get_target_from_rule(module, client, rule, name, group, groups, vpc_id, tags):
    """
    Returns tuple of (target_type, target, group_created) after validating rule params.

    rule: Dict describing a rule.
    name: Name of the security group being managed.
    groups: Dict of all available security groups.

    AWS accepts an ip range or a security group as target of a rule. This
    function validate the rule specification and return either a non-None
    group_id or a non-None ip range.

    When using a security group as a target all 3 fields (OwnerId, GroupId, and
    GroupName) need to exist in the target. This ensures consistency of the
    values that will be compared to current_rules (from current_ingress and
    current_egress) in wait_for_rule_propagation().
    """
    try:
        if rule.get("group_id"):
            return _target_from_rule_with_group_id(rule, groups)
        if "group_name" in rule:
            return _target_from_rule_with_group_name(client, rule, name, group, groups, vpc_id, tags, module.check_mode)
        if "cidr_ip" in rule:
            return "ipv4", validate_ip(module, rule["cidr_ip"]), False
        if "cidr_ipv6" in rule:
            return "ipv6", validate_ip(module, rule["cidr_ipv6"]), False
        if "ip_prefix" in rule:
            return "ip_prefix", rule["ip_prefix"], False
    except SecurityGroupError as e:
        e.fail(module)

    module.fail_json(msg="Could not match target for rule", failed_rule=rule)


def _strip_rule(rule):
    """
    Returns a copy of the rule with the Target/Source and Port information
    from a rule stripped out.
    This can then be combined with the expanded information
    """
    stripped_rule = deepcopy(rule)
    # Get just the non-source/port info from the rule
    [stripped_rule.pop(source_type, None) for source_type in SOURCE_TYPES_ALL]
    [stripped_rule.pop(port_type, None) for port_type in PORT_TYPES_ALL]
    return stripped_rule


def expand_rules(rules):
    if rules is None:
        return rules

    expanded_rules = []
    for rule in rules:
        expanded_rules.extend(expand_rule(rule))

    return expanded_rules


def expand_rule(rule):
    rule = scrub_none_parameters(rule)
    ports_list = expand_ports_from_rule(rule)
    sources_list = expand_sources_from_rule(rule)
    stripped_rule = _strip_rule(rule)

    # expands out all possible combinations of ports and sources for the rule
    # This results in a list of pairs of dictionaries...
    ports_and_sources = itertools.product(ports_list, sources_list)

    # Combines each pair of port/source dictionaries with rest of the info from the rule
    return [{**stripped_rule, **port, **source} for (port, source) in ports_and_sources]


def expand_sources_from_rule(rule):
    sources = []
    for type_name in sorted(SOURCE_TYPES_ALL):
        if rule.get(type_name) is not None:
            sources.extend([{type_name: target} for target in rule.get(type_name)])
    if not sources:
        raise SecurityGroupError("Unable to find source/target information in rule", rule=rule)
    return tuple(sources)


def expand_ports_from_rule(rule):
    # While icmp_type/icmp_code could have been aliases, this wouldn't be obvious in the
    # documentation
    if rule.get("icmp_type") is not None:
        return ({"from_port": rule.get("icmp_type"), "to_port": rule.get("icmp_code")},)
    if rule.get("from_port") is not None or rule.get("to_port") is not None:
        return ({"from_port": rule.get("from_port"), "to_port": rule.get("to_port")},)
    if rule.get("ports") is not None:
        ports = expand_ports_list(rule.get("ports"))
        return tuple({"from_port": from_port, "to_port": to_port} for (from_port, to_port) in ports)
    return ({},)


def expand_ports_list(ports):
    # takes a list of ports and returns a list of (port_from, port_to)
    ports_expanded = []
    for port in ports:
        try:
            port_list = (int(port.strip()),) * 2
        except ValueError as e:
            # Someone passed a range
            if "-" in port:
                port_list = [int(p.strip()) for p in port.split("-", 1)]
            else:
                raise SecurityGroupError("Unable to parse port", port=port) from e
        ports_expanded.append(tuple(sorted(port_list)))

    return ports_expanded


def update_rules_description(module, client, rule_type, group_id, ip_permissions):
    if module.check_mode:
        return
    try:
        if rule_type == "in":
            client.update_security_group_rule_descriptions_ingress(
                aws_retry=True, GroupId=group_id, IpPermissions=ip_permissions
            )
        if rule_type == "out":
            client.update_security_group_rule_descriptions_egress(
                aws_retry=True, GroupId=group_id, IpPermissions=ip_permissions
            )
    except (ClientError, BotoCoreError) as e:
        module.fail_json_aws(e, msg=f"Unable to update rule description for group {group_id}")


def fix_port_and_protocol(permission):
    for key in ("FromPort", "ToPort"):
        if key in permission:
            if permission[key] is None:
                del permission[key]
            else:
                permission[key] = int(permission[key])

    permission["IpProtocol"] = to_text(permission["IpProtocol"])

    return permission


def remove_old_permissions(client, module, revoke_ingress, revoke_egress, group_id):
    if revoke_ingress:
        revoke(client, module, revoke_ingress, group_id, "in")
    if revoke_egress:
        revoke(client, module, revoke_egress, group_id, "out")
    return bool(revoke_ingress or revoke_egress)


def revoke(client, module, ip_permissions, group_id, rule_type):
    if not module.check_mode:
        try:
            if rule_type == "in":
                client.revoke_security_group_ingress(aws_retry=True, GroupId=group_id, IpPermissions=ip_permissions)
            elif rule_type == "out":
                client.revoke_security_group_egress(aws_retry=True, GroupId=group_id, IpPermissions=ip_permissions)
        except (BotoCoreError, ClientError) as e:
            rules = "ingress rules" if rule_type == "in" else "egress rules"
            module.fail_json_aws(e, f"Unable to revoke {rules}: {ip_permissions}")


def add_new_permissions(client, module, new_ingress, new_egress, group_id):
    if new_ingress:
        authorize(client, module, new_ingress, group_id, "in")
    if new_egress:
        authorize(client, module, new_egress, group_id, "out")
    return bool(new_ingress or new_egress)


def authorize(client, module, ip_permissions, group_id, rule_type):
    if not module.check_mode:
        try:
            if rule_type == "in":
                client.authorize_security_group_ingress(aws_retry=True, GroupId=group_id, IpPermissions=ip_permissions)
            elif rule_type == "out":
                client.authorize_security_group_egress(aws_retry=True, GroupId=group_id, IpPermissions=ip_permissions)
        except (BotoCoreError, ClientError) as e:
            rules = "ingress rules" if rule_type == "in" else "egress rules"
            module.fail_json_aws(e, f"Unable to authorize {rules}: {ip_permissions}")


def validate_ip(module, cidr_ip):
    split_addr = cidr_ip.split("/")
    if len(split_addr) != 2:
        return cidr_ip

    try:
        ip = ip_network(to_text(cidr_ip))
        return str(ip)
    except ValueError:
        # If a host bit is incorrectly set, ip_network will throw an error at us,
        # we'll continue, convert the address to a CIDR AWS will accept, but issue a warning.
        pass

    # Try evaluating as an IPv4 network, it'll throw a ValueError if it can't parse cidr_ip as an
    # IPv4 network
    try:
        ip = to_subnet(split_addr[0], split_addr[1])
        module.warn(
            f"One of your CIDR addresses ({cidr_ip}) has host bits set. To get rid of this warning, check the network"
            f" mask and make sure that only network bits are set: {ip}."
        )
        return ip
    except ValueError:
        pass

    # Try again, evaluating as an IPv6 network.
    try:
        ip6 = to_ipv6_subnet(split_addr[0]) + "/" + split_addr[1]
        module.warn(
            f"One of your IPv6 CIDR addresses ({cidr_ip}) has host bits set. To get rid of this warning, check the"
            f" network mask and make sure that only network bits are set: {ip6}."
        )
        return ip6
    except ValueError:
        module.warn(f"Unable to parse CIDR ({cidr_ip}).")
        return cidr_ip


def update_rule_descriptions(
    module, client, group_id, present_ingress, named_tuple_ingress_list, present_egress, named_tuple_egress_list
):
    changed = False
    ingress_needs_desc_update = []
    egress_needs_desc_update = []

    for present_rule in present_egress:
        needs_update = [
            r
            for r in named_tuple_egress_list
            if rule_cmp(r, present_rule) and r.description != present_rule.description
        ]
        for r in needs_update:
            named_tuple_egress_list.remove(r)
        egress_needs_desc_update.extend(needs_update)
    for present_rule in present_ingress:
        needs_update = [
            r
            for r in named_tuple_ingress_list
            if rule_cmp(r, present_rule) and r.description != present_rule.description
        ]
        for r in needs_update:
            named_tuple_ingress_list.remove(r)
        ingress_needs_desc_update.extend(needs_update)

    if ingress_needs_desc_update:
        update_rules_description(module, client, "in", group_id, rules_to_permissions(ingress_needs_desc_update))
        changed |= True
    if egress_needs_desc_update:
        update_rules_description(module, client, "out", group_id, rules_to_permissions(egress_needs_desc_update))
        changed |= True
    return changed


def _create_security_group_with_wait(client, name, description, vpc_id, tags):
    params = dict(GroupName=name, Description=description)
    if vpc_id:
        params["VpcId"] = vpc_id
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, ["security-group"])

    created_group = client.create_security_group(aws_retry=True, **params)
    get_waiter(
        client,
        "security_group_exists",
    ).wait(
        GroupIds=[created_group["GroupId"]],
    )
    return created_group


def create_security_group(client, module, name, description, vpc_id, tags):
    if not module.check_mode:
        params = dict(GroupName=name, Description=description)
        if vpc_id:
            params["VpcId"] = vpc_id
        if tags:
            params["TagSpecifications"] = boto3_tag_specifications(tags, ["security-group"])
        try:
            group = client.create_security_group(aws_retry=True, **params)
        except (BotoCoreError, ClientError) as e:
            module.fail_json_aws(e, msg="Unable to create security group")
        # When a group is created, an egress_rule ALLOW ALL
        # to 0.0.0.0/0 is added automatically but it's not
        # reflected in the object returned by the AWS API
        # call. We re-read the group for getting an updated object
        # amazon sometimes takes a couple seconds to update the security group so wait till it exists
        while True:
            sleep(3)
            group = get_security_groups_with_backoff(client, GroupIds=[group["GroupId"]])["SecurityGroups"][0]
            if group.get("VpcId") and not group.get("IpPermissionsEgress"):
                pass
            else:
                break
        return group
    return None


def wait_for_rule_propagation(module, client, group, desired_ingress, desired_egress, purge_ingress, purge_egress):
    group_id = group["GroupId"]
    tries = 6

    def await_rules(group, desired_rules, purge, rule_key):
        for _i in range(tries):
            current_rules = set(sum([list(rule_from_group_permission(p)) for p in group[rule_key]], []))
            if purge and len(current_rules ^ set(desired_rules)) == 0:
                return group
            elif purge:
                conflicts = current_rules ^ set(desired_rules)
                # For cases where set comparison is equivalent, but invalid port/proto exist
                for a, b in itertools.combinations(conflicts, 2):
                    if rule_cmp(a, b):
                        conflicts.discard(a)
                        conflicts.discard(b)
                if not len(conflicts):
                    return group
            elif current_rules.issuperset(desired_rules) and not purge:
                return group
            sleep(10)
            group = get_security_groups_with_backoff(client, GroupIds=[group_id])["SecurityGroups"][0]
        module.warn(
            f"Ran out of time waiting for {group_id} {rule_key}. Current: {current_rules}, Desired: {desired_rules}"
        )
        return group

    group = get_security_groups_with_backoff(client, GroupIds=[group_id])["SecurityGroups"][0]
    if "VpcId" in group and module.params.get("rules_egress") is not None:
        group = await_rules(group, desired_egress, purge_egress, "IpPermissionsEgress")
    return await_rules(group, desired_ingress, purge_ingress, "IpPermissions")


def group_exists(client, module, vpc_id, group_id, name):
    filters = dict()
    params = dict()
    if group_id:
        if isinstance(group_id, list):
            params["GroupIds"] = group_id
        else:
            params["GroupIds"] = [group_id]
    if name:
        # Add name to filters rather than params['GroupNames']
        # because params['GroupNames'] only checks the default vpc if no vpc is provided
        filters["group-name"] = name
    if vpc_id:
        filters["vpc-id"] = vpc_id
    # Don't filter by description to maintain backwards compatibility
    params["Filters"] = ansible_dict_to_boto3_filter_list(filters)
    try:
        security_groups = sg_exists_with_backoff(client, **params).get("SecurityGroups", [])
        all_groups = get_security_groups_with_backoff(client).get("SecurityGroups", [])
    except (BotoCoreError, ClientError) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Error in describe_security_groups")

    if security_groups:
        groups = dict((group["GroupId"], group) for group in all_groups)
        groups.update(dict((group["GroupName"], group) for group in all_groups))
        if vpc_id:
            vpc_wins = dict(
                (group["GroupName"], group) for group in all_groups if group.get("VpcId") and group["VpcId"] == vpc_id
            )
            groups.update(vpc_wins)
        # maintain backwards compatibility by using the last matching group
        return security_groups[-1], groups
    return None, {}


def get_diff_final_resource(client, module, security_group):
    def get_account_id(security_group, module):
        try:
            owner_id = security_group.get("owner_id", current_account_id)
        except (BotoCoreError, ClientError) as e:
            owner_id = f"Unable to determine owner_id: {to_text(e)}"
        return owner_id

    def get_final_tags(security_group_tags, specified_tags, purge_tags):
        if specified_tags is None:
            return security_group_tags
        tags_need_modify, tags_to_delete = compare_aws_tags(security_group_tags, specified_tags, purge_tags)
        end_result_tags = dict((k, v) for k, v in specified_tags.items() if k not in tags_to_delete)
        end_result_tags.update(dict((k, v) for k, v in security_group_tags.items() if k not in tags_to_delete))
        end_result_tags.update(tags_need_modify)
        return end_result_tags

    def get_final_rules(client, module, security_group_rules, specified_rules, purge_rules):
        if specified_rules is None:
            return security_group_rules
        if purge_rules:
            final_rules = []
        else:
            final_rules = list(security_group_rules)
        specified_rules = flatten_nested_targets(module, deepcopy(specified_rules))
        for rule in specified_rules:
            format_rule = {
                "from_port": None,
                "to_port": None,
                "ip_protocol": rule.get("proto"),
                "ip_ranges": [],
                "ipv6_ranges": [],
                "prefix_list_ids": [],
                "user_id_group_pairs": [],
            }
            if rule.get("proto") in ("all", "-1", -1):
                format_rule["ip_protocol"] = "-1"
                format_rule.pop("from_port")
                format_rule.pop("to_port")
            elif rule.get("ports"):
                if rule.get("ports") and (isinstance(rule["ports"], string_types) or isinstance(rule["ports"], int)):
                    rule["ports"] = [rule["ports"]]
                for port in rule.get("ports"):
                    if isinstance(port, string_types) and "-" in port:
                        format_rule["from_port"], format_rule["to_port"] = port.split("-")
                    else:
                        format_rule["from_port"] = format_rule["to_port"] = port
            elif rule.get("from_port") or rule.get("to_port"):
                format_rule["from_port"] = rule.get("from_port", rule.get("to_port"))
                format_rule["to_port"] = rule.get("to_port", rule.get("from_port"))
            for source_type in ("cidr_ip", "cidr_ipv6", "prefix_list_id"):
                if rule.get(source_type):
                    rule_key = {
                        "cidr_ip": "ip_ranges",
                        "cidr_ipv6": "ipv6_ranges",
                        "prefix_list_id": "prefix_list_ids",
                    }.get(source_type)
                    if rule.get("rule_desc"):
                        format_rule[rule_key] = [{source_type: rule[source_type], "description": rule["rule_desc"]}]
                    else:
                        if not isinstance(rule[source_type], list):
                            rule[source_type] = [rule[source_type]]
                        format_rule[rule_key] = [{source_type: target} for target in rule[source_type]]
            if rule.get("group_id") or rule.get("group_name"):
                # XXX bug - doesn't cope with a list of ids/names
                rule_sg = group_exists(
                    client, module, module.params["vpc_id"], rule.get("group_id"), rule.get("group_name")
                )[0]
                if rule_sg is None:
                    # --diff during --check
                    format_rule["user_id_group_pairs"] = [
                        {
                            "group_id": rule.get("group_id"),
                            "group_name": rule.get("group_name"),
                            "peering_status": None,
                            "user_id": get_account_id(security_group, module),
                            "vpc_id": module.params["vpc_id"],
                            "vpc_peering_connection_id": None,
                        }
                    ]
                else:
                    rule_sg = camel_dict_to_snake_dict(rule_sg)
                    format_rule["user_id_group_pairs"] = [
                        {
                            "description": rule_sg.get("description", rule_sg.get("group_desc")),
                            "group_id": rule_sg.get("group_id", rule.get("group_id")),
                            "group_name": rule_sg.get("group_name", rule.get("group_name")),
                            "peering_status": rule_sg.get("peering_status"),
                            "user_id": rule_sg.get("user_id", get_account_id(security_group, module)),
                            "vpc_id": rule_sg.get("vpc_id", module.params["vpc_id"]),
                            "vpc_peering_connection_id": rule_sg.get("vpc_peering_connection_id"),
                        }
                    ]
                for k, v in list(format_rule["user_id_group_pairs"][0].items()):
                    if v is None:
                        format_rule["user_id_group_pairs"][0].pop(k)
            final_rules.append(format_rule)
        return final_rules

    security_group_ingress = security_group.get("ip_permissions", [])
    specified_ingress = module.params["rules"]
    purge_ingress = module.params["purge_rules"]
    security_group_egress = security_group.get("ip_permissions_egress", [])
    specified_egress = module.params["rules_egress"]
    purge_egress = module.params["purge_rules_egress"]
    return {
        "description": module.params["description"],
        "group_id": security_group.get("group_id", "sg-xxxxxxxx"),
        "group_name": security_group.get("group_name", module.params["name"]),
        "ip_permissions": get_final_rules(client, module, security_group_ingress, specified_ingress, purge_ingress),
        "ip_permissions_egress": get_final_rules(client, module, security_group_egress, specified_egress, purge_egress),
        "owner_id": get_account_id(security_group, module),
        "tags": get_final_tags(security_group.get("tags", {}), module.params["tags"], module.params["purge_tags"]),
        "vpc_id": security_group.get("vpc_id", module.params["vpc_id"]),
    }


def flatten_nested_targets(module, rules):
    def _flatten(targets):
        for target in targets:
            if isinstance(target, list):
                module.deprecate(
                    (
                        "Support for nested lists in cidr_ip and cidr_ipv6 has been "
                        "deprecated.  The flatten filter can be used instead."
                    ),
                    date="2024-12-01",
                    collection_name="amazon.aws",
                )
                for t in _flatten(target):
                    yield t
            elif isinstance(target, string_types):
                yield target

    if rules is not None:
        for rule in rules:
            target_list_type = None
            if isinstance(rule.get("cidr_ip"), list):
                target_list_type = "cidr_ip"
            elif isinstance(rule.get("cidr_ipv6"), list):
                target_list_type = "cidr_ipv6"
            if target_list_type is not None:
                rule[target_list_type] = list(_flatten(rule[target_list_type]))
    return rules


def get_rule_sort_key(dicts):
    if dicts.get("cidr_ip"):
        return str(dicts.get("cidr_ip"))
    if dicts.get("cidr_ipv6"):
        return str(dicts.get("cidr_ipv6"))
    if dicts.get("prefix_list_id"):
        return str(dicts.get("prefix_list_id"))
    if dicts.get("group_id"):
        return str(dicts.get("group_id"))
    return None


def get_ip_permissions_sort_key(rule):
    RULE_KEYS_ALL = {"ip_ranges", "ipv6_ranges", "prefix_list_ids", "user_id_group_pairs"}
    # Ensure content of these keys is sorted
    for rule_key in RULE_KEYS_ALL:
        if rule.get(rule_key):
            rule.get(rule_key).sort(key=get_rule_sort_key)

    # Returns the first value plus a prefix so the types get clustered together when sorted
    if rule.get("ip_ranges"):
        value = str(rule.get("ip_ranges")[0]["cidr_ip"])
        return f"ipv4:{value}"
    if rule.get("ipv6_ranges"):
        value = str(rule.get("ipv6_ranges")[0]["cidr_ipv6"])
        return f"ipv6:{value}"
    if rule.get("prefix_list_ids"):
        value = str(rule.get("prefix_list_ids")[0]["prefix_list_id"])
        return f"pl:{value}"
    if rule.get("user_id_group_pairs"):
        value = str(rule.get("user_id_group_pairs")[0].get("group_id", ""))
        return f"ugid:{value}"
    return None


def sort_security_group(security_group):
    if not security_group:
        return security_group

    if security_group.get("ip_permissions"):
        security_group["ip_permissions"].sort(key=get_ip_permissions_sort_key)
    if security_group.get("ip_permissions_egress"):
        security_group["ip_permissions_egress"].sort(key=get_ip_permissions_sort_key)

    return security_group


def validate_rules(module, rules):
    if not rules:
        return
    try:
        for rule in rules:
            validate_rule(rule)
    except SecurityGroupError as e:
        e.fail(module)


def ensure_absent(client, group, check_mode):
    if not group:
        return False
    if check_mode:
        return True

    try:
        client.delete_security_group(aws_retry=True, GroupId=group["GroupId"])
    except is_boto3_error_code("InvalidGroup.NotFound"):
        return False
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        raise SecurityGroupError(f"Unable to delete security group '{group}'", e=e)

    return True


def ensure_present(module, client, group, groups):
    name = module.params["name"]
    group_id = module.params["group_id"]
    description = module.params["description"]
    vpc_id = module.params["vpc_id"]
    # Deprecated
    rules = flatten_nested_targets(module, deepcopy(module.params["rules"]))
    rules_egress = flatten_nested_targets(module, deepcopy(module.params["rules_egress"]))
    # /end Deprecated
    validate_rules(module, rules)
    validate_rules(module, rules_egress)
    rules = deduplicate_rules_args(expand_rules(rules))
    rules_egress = deduplicate_rules_args(expand_rules(rules_egress))
    state = module.params.get("state")
    purge_rules = module.params["purge_rules"]
    purge_rules_egress = module.params["purge_rules_egress"]
    tags = module.params["tags"]
    purge_tags = module.params["purge_tags"]

    changed = False
    group_created_new = False

    if not group:
        # Short circuit things if we're in check_mode
        if module.check_mode:
            return True, None

        group = create_security_group(client, module, name, description, vpc_id, tags)
        group_created_new = True
        changed = True

    else:
        # Description is immutable
        if group["Description"] != description:
            module.warn(
                "Group description does not match existing group. Descriptions cannot be changed without deleting "
                "and re-creating the security group. Try using state=absent to delete, then rerunning this task."
            )

    changed |= ensure_ec2_tags(client, module, group["GroupId"], tags=tags, purge_tags=purge_tags)

    named_tuple_ingress_list = []
    named_tuple_egress_list = []
    current_ingress = sum([list(rule_from_group_permission(p)) for p in group["IpPermissions"]], [])
    current_egress = sum([list(rule_from_group_permission(p)) for p in group["IpPermissionsEgress"]], [])

    for new_rules, _rule_type, named_tuple_rule_list in [
        (rules, "in", named_tuple_ingress_list),
        (rules_egress, "out", named_tuple_egress_list),
    ]:
        if new_rules is None:
            continue
        for rule in new_rules:
            target_type, target, target_group_created = get_target_from_rule(
                module, client, rule, name, group, groups, vpc_id, tags
            )
            changed |= target_group_created

            if rule.get("proto") in ("all", "-1", -1):
                rule["proto"] = "-1"
                rule["from_port"] = None
                rule["to_port"] = None

            try:
                int(rule.get("proto"))
                rule["proto"] = to_text(rule.get("proto"))
                rule["from_port"] = None
                rule["to_port"] = None
            except ValueError:
                # rule does not use numeric protocol spec
                pass
            named_tuple_rule_list.append(
                Rule(
                    port_range=(rule["from_port"], rule["to_port"]),
                    protocol=to_text(rule.get("proto")),
                    target=target,
                    target_type=target_type,
                    description=rule.get("rule_desc"),
                )
            )

    # List comprehensions for rules to add, rules to modify, and rule ids to determine purging
    new_ingress_permissions = [to_permission(r) for r in (set(named_tuple_ingress_list) - set(current_ingress))]
    new_egress_permissions = [to_permission(r) for r in (set(named_tuple_egress_list) - set(current_egress))]

    if module.params.get("rules_egress") is None and "VpcId" in group:
        # when no egress rules are specified and we're in a VPC,
        # we add in a default allow all out rule, which was the
        # default behavior before egress rules were added
        rule = Rule((None, None), "-1", "0.0.0.0/0", "ipv4", None)
        if rule in current_egress:
            named_tuple_egress_list.append(rule)
        if rule not in current_egress:
            current_egress.append(rule)

    # List comprehensions for rules to add, rules to modify, and rule ids to determine purging
    present_ingress = list(set(named_tuple_ingress_list).union(set(current_ingress)))
    present_egress = list(set(named_tuple_egress_list).union(set(current_egress)))

    if purge_rules:
        revoke_ingress = []
        for p in present_ingress:
            if not any(rule_cmp(p, b) for b in named_tuple_ingress_list):
                revoke_ingress.append(to_permission(p))
    else:
        revoke_ingress = []

    if purge_rules_egress and module.params.get("rules_egress") is not None:
        revoke_egress = []
        for p in present_egress:
            if not any(rule_cmp(p, b) for b in named_tuple_egress_list):
                revoke_egress.append(to_permission(p))
    else:
        revoke_egress = []

    # named_tuple_ingress_list and named_tuple_egress_list get updated by
    # method update_rule_descriptions, deep copy these two lists to new
    # variables for the record of the 'desired' ingress and egress sg permissions
    desired_ingress = deepcopy(named_tuple_ingress_list)
    desired_egress = deepcopy(named_tuple_egress_list)

    changed |= update_rule_descriptions(
        module,
        client,
        group["GroupId"],
        present_ingress,
        named_tuple_ingress_list,
        present_egress,
        named_tuple_egress_list,
    )

    # Revoke old rules
    changed |= remove_old_permissions(client, module, revoke_ingress, revoke_egress, group["GroupId"])

    new_ingress_permissions = [to_permission(r) for r in (set(named_tuple_ingress_list) - set(current_ingress))]
    new_ingress_permissions = rules_to_permissions(set(named_tuple_ingress_list) - set(current_ingress))
    new_egress_permissions = rules_to_permissions(set(named_tuple_egress_list) - set(current_egress))
    # Authorize new rules
    changed |= add_new_permissions(client, module, new_ingress_permissions, new_egress_permissions, group["GroupId"])

    if group_created_new and module.params.get("rules") is None and module.params.get("rules_egress") is None:
        # A new group with no rules provided is already being awaited.
        # When it is created we wait for the default egress rule to be added by AWS
        security_group = get_security_groups_with_backoff(client, GroupIds=[group["GroupId"]])["SecurityGroups"][0]
    elif changed and not module.check_mode:
        # keep pulling until current security group rules match the desired ingress and egress rules
        security_group = wait_for_rule_propagation(
            module, client, group, desired_ingress, desired_egress, purge_rules, purge_rules_egress
        )
    else:
        security_group = get_security_groups_with_backoff(client, GroupIds=[group["GroupId"]])["SecurityGroups"][0]
    security_group = camel_dict_to_snake_dict(security_group, ignore_list=["Tags"])
    security_group["tags"] = boto3_tag_list_to_ansible_dict(security_group.get("tags", []))

    return changed, security_group


def main():
    rule_spec = dict(
        rule_desc=dict(type="str"),
        # We have historically allowed for lists of lists in cidr_ip and cidr_ipv6
        # https://github.com/ansible-collections/amazon.aws/pull/1213
        cidr_ip=dict(type="list", elements="raw"),
        cidr_ipv6=dict(type="list", elements="raw"),
        ip_prefix=dict(type="list", elements="str"),
        group_id=dict(type="list", elements="str"),
        group_name=dict(type="list", elements="str"),
        group_desc=dict(type="str"),
        proto=dict(type="str", default="tcp"),
        ports=dict(type="list", elements="str"),
        from_port=dict(type="int"),
        to_port=dict(type="int"),
        icmp_type=dict(type="int"),
        icmp_code=dict(type="int"),
    )
    rule_requirements = dict(
        mutually_exclusive=(
            # PORTS / ICMP_TYPE + ICMP_CODE / TO_PORT + FROM_PORT
            (
                "ports",
                "to_port",
            ),
            (
                "ports",
                "from_port",
            ),
            (
                "ports",
                "icmp_type",
            ),
            (
                "ports",
                "icmp_code",
            ),
            (
                "icmp_type",
                "to_port",
            ),
            (
                "icmp_code",
                "to_port",
            ),
            (
                "icmp_type",
                "from_port",
            ),
            (
                "icmp_code",
                "from_port",
            ),
        ),
        required_one_of=(
            # A target must be specified
            (
                "group_id",
                "group_name",
                "cidr_ip",
                "cidr_ipv6",
                "ip_prefix",
            ),
        ),
        required_by=dict(
            # If you specify an ICMP code, you must specify the ICMP type
            icmp_code=("icmp_type",),
        ),
    )

    argument_spec = dict(
        name=dict(),
        group_id=dict(),
        description=dict(),
        vpc_id=dict(),
        rules=dict(type="list", elements="dict", options=rule_spec, **rule_requirements),
        rules_egress=dict(
            type="list", elements="dict", aliases=["egress_rules"], options=rule_spec, **rule_requirements
        ),
        state=dict(default="present", type="str", choices=["present", "absent"]),
        purge_rules=dict(default=True, required=False, type="bool"),
        purge_rules_egress=dict(default=True, required=False, type="bool", aliases=["purge_egress_rules"]),
        tags=dict(required=False, type="dict", aliases=["resource_tags"]),
        purge_tags=dict(default=True, required=False, type="bool"),
    )
    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=[["name", "group_id"]],
        required_if=[["state", "present", ["name", "description"]]],
    )

    name = module.params["name"]
    group_id = module.params["group_id"]
    vpc_id = module.params["vpc_id"]
    state = module.params.get("state")

    client = module.client("ec2", AWSRetry.jittered_backoff())

    group, groups = group_exists(client, module, vpc_id, group_id, name)

    global current_account_id
    current_account_id = get_aws_account_id(module)

    before = {}
    after = {}

    if group:
        before = camel_dict_to_snake_dict(group, ignore_list=["Tags"])
        before["tags"] = boto3_tag_list_to_ansible_dict(before.get("tags", []))

    try:
        # Ensure requested group is absent
        if state == "absent":
            changed = ensure_absent(client, group, module.check_mode)
            security_group = {"group_id": None}
        # Ensure requested group is present
        elif state == "present":
            (changed, security_group) = ensure_present(module, client, group, groups)
            # Check mode can't create anything
            if not security_group:
                security_group = {"group_id": None}
    except SecurityGroupError as e:
        e.fail(module)

    if module._diff:
        if state == "present":
            after = get_diff_final_resource(client, module, security_group)

        # Order final rules consistently
        before = sort_security_group(before)
        after = sort_security_group(after)

        security_group["diff"] = [{"before": before, "after": after}]

    module.exit_json(changed=changed, **security_group)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: discovery_device_range_info
short_description: Information module for Discovery Device Range
description:
- Get all Discovery Device Range.
- >
   Returns the network devices discovered for the given discovery and for the given range. The maximum number of
   records that can be retrieved is 500. Discovery ID can be obtained using the "Get Discoveries by range" API.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  id:
    description:
    - Id path parameter. Discovery ID.
    type: str
  startIndex:
    description:
    - StartIndex path parameter. Start index.
    type: int
  recordsToReturn:
    description:
    - RecordsToReturn path parameter. Number of records to return.
    type: int
  taskId:
    description:
    - TaskId query parameter.
    type: str
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Discovery GetDiscoveredDevicesByRange
  description: Complete reference of the GetDiscoveredDevicesByRange API.
  link: https://developer.cisco.com/docs/dna-center/#!get-discovered-devices-by-range
notes:
  - SDK Method used are
    discovery.Discovery.get_discovered_devices_by_range,

  - Paths used are
    get /dna/intent/api/v1/discovery/{id}/network-device/{startIndex}/{recordsToReturn},

"""

EXAMPLES = r"""
- name: Get all Discovery Device Range
  cisco.dnac.discovery_device_range_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    taskId: string
    id: string
    startIndex: 0
    recordsToReturn: 0
  register: result

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": [
        {
          "anchorWlcForAp": "string",
          "authModelId": "string",
          "avgUpdateFrequency": 0,
          "bootDateTime": "string",
          "cliStatus": "string",
          "duplicateDeviceId": "string",
          "errorCode": "string",
          "errorDescription": "string",
          "family": "string",
          "hostname": "string",
          "httpStatus": "string",
          "id": "string",
          "imageName": "string",
          "ingressQueueConfig": "string",
          "interfaceCount": "string",
          "inventoryCollectionStatus": "string",
          "inventoryReachabilityStatus": "string",
          "lastUpdated": "string",
          "lineCardCount": "string",
          "lineCardId": "string",
          "location": "string",
          "locationName": "string",
          "macAddress": "string",
          "managementIpAddress": "string",
          "memorySize": "string",
          "netconfStatus": "string",
          "numUpdates": 0,
          "pingStatus": "string",
          "platformId": "string",
          "portRange": "string",
          "qosStatus": "string",
          "reachabilityFailureReason": "string",
          "reachabilityStatus": "string",
          "role": "string",
          "roleSource": "string",
          "serialNumber": "string",
          "snmpContact": "string",
          "snmpLocation": "string",
          "snmpStatus": "string",
          "softwareVersion": "string",
          "tag": "string",
          "tagCount": 0,
          "type": "string",
          "upTime": "string",
          "vendor": "string",
          "wlcApDeviceStatus": "string"
        }
      ],
      "version": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: s3_lifecycle
version_added: 1.0.0
short_description: Manage S3 bucket lifecycle rules in AWS
description:
  - Manage S3 bucket lifecycle rules in AWS.
author:
  - "Rob White (@wimnat)"
notes:
  - If specifying expiration time as days then transition time must also be specified in days.
  - If specifying expiration time as a date then transition time must also be specified as a date.
options:
  name:
    description:
      - Name of the S3 bucket.
    required: true
    type: str
  abort_incomplete_multipart_upload_days:
    description:
      - Specifies the days since the initiation of an incomplete multipart upload that Amazon S3 will wait before permanently removing all parts of the upload.
    type: int
    version_added: 2.2.0
  expiration_date:
    description:
      - Indicates the lifetime of the objects that are subject to the rule by the date they will expire.
      - The value must be ISO-8601 format, the time must be midnight and a GMT timezone must be specified.
      - This cannot be specified with I(expire_object_delete_marker)
    type: str
  expiration_days:
    description:
      - Indicates the lifetime, in days, of the objects that are subject to the rule.
      - The value must be a non-zero positive integer.
      - This cannot be specified with I(expire_object_delete_marker)
    type: int
  expire_object_delete_marker:
    description:
      - Indicates whether Amazon S3 will remove a delete marker with no noncurrent versions.
      - If set to C(true), the delete marker will be expired; if set to C(false) the policy takes no action.
      - This cannot be specified with I(expiration_days) or I(expiration_date).
    type: bool
    version_added: 2.2.0
  prefix:
    description:
      - Prefix identifying one or more objects to which the rule applies.
      - If no prefix is specified, the rule will apply to the whole bucket.
    type: str
  purge_transitions:
    description:
      - Whether to replace all the current transition(s) with the new transition(s).
      - When C(false), the provided transition(s) will be added, replacing transitions
        with the same storage_class. When true, existing transitions will be removed
        and replaced with the new transition(s)
    default: true
    type: bool
  noncurrent_version_expiration_days:
    description:
      - The number of days after which non-current versions should be deleted.
      - Must be set if I(noncurrent_version_keep_newer) is set.
    required: false
    type: int
  noncurrent_version_keep_newer:
    description:
      - The minimum number of non-current versions to retain.
      - Requres I(noncurrent_version_expiration_days).
    required: false
    type: int
    version_added: 5.3.0
  noncurrent_version_storage_class:
    description:
      - The storage class to which non-current versions are transitioned.
    default: glacier
    choices: ['glacier', 'onezone_ia', 'standard_ia', 'intelligent_tiering', 'deep_archive']
    required: false
    type: str
  noncurrent_version_transition_days:
    description:
      - The number of days after which non-current versions will be transitioned
        to the storage class specified in I(noncurrent_version_storage_class).
    required: false
    type: int
  noncurrent_version_transitions:
    description:
      - A list of transition behaviors to be applied to noncurrent versions for the rule.
      - Each storage class may be used only once. Each transition behavior contains these elements
          I(transition_days)
          I(storage_class)
    type: list
    elements: dict
  rule_id:
    description:
      - Unique identifier for the rule.
      - The value cannot be longer than 255 characters.
      - A unique value for the rule will be generated if no value is provided.
    type: str
  state:
    description:
      - Create or remove the lifecycle rule.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  status:
    description:
      - If C(enabled), the rule is currently being applied.
      - If C(disabled), the rule is not currently being applied.
    default: enabled
    choices: [ 'enabled', 'disabled' ]
    type: str
  storage_class:
    description:
      - The storage class to transition to.
    default: glacier
    choices: [ 'glacier', 'onezone_ia', 'standard_ia', 'intelligent_tiering', 'deep_archive']
    type: str
  transition_date:
    description:
      - Indicates the lifetime of the objects that are subject to the rule by the date they
        will transition to a different storage class.
      - The value must be ISO-8601 format, the time must be midnight and a GMT timezone must
        be specified.
      - If (transition_days) is not specified, this parameter is required.
    type: str
  transition_days:
    description:
      - Indicates when, in days, an object transitions to a different storage class.
      - If I(transition_date) is not specified, this parameter is required.
    type: int
  transitions:
    description:
      - A list of transition behaviors to be applied to the rule.
      - Each storage class may be used only once. Each transition behavior may contain these elements
          I(transition_days)
          I(transition_date)
          I(storage_class)
    type: list
    elements: dict
  wait:
    description:
      - Wait for the configuration to complete before returning.
    version_added: 1.5.0
    type: bool
    default: false
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

RETURN = r""" # """

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Configure a lifecycle rule on a bucket to expire (delete) items with a prefix of /logs/ after 30 days
  community.aws.s3_lifecycle:
    name: mybucket
    expiration_days: 30
    prefix: logs/
    status: enabled
    state: present

- name: Configure a lifecycle rule to transition all items with a prefix of /logs/ to glacier after 7 days and then delete after 90 days
  community.aws.s3_lifecycle:
    name: mybucket
    transition_days: 7
    expiration_days: 90
    prefix: logs/
    status: enabled
    state: present

# Note that midnight GMT must be specified.
# Be sure to quote your date strings
- name: Configure a lifecycle rule to transition all items with a prefix of /logs/ to glacier on 31 Dec 2020 and then delete on 31 Dec 2030.
  community.aws.s3_lifecycle:
    name: mybucket
    transition_date: "2020-12-30T00:00:00.000Z"
    expiration_date: "2030-12-30T00:00:00.000Z"
    prefix: logs/
    status: enabled
    state: present

- name: Disable the rule created above
  community.aws.s3_lifecycle:
    name: mybucket
    prefix: logs/
    status: disabled
    state: present

- name: Delete the lifecycle rule created above
  community.aws.s3_lifecycle:
    name: mybucket
    prefix: logs/
    state: absent

- name: Configure a lifecycle rule to transition all backup files older than 31 days in /backups/ to standard infrequent access class.
  community.aws.s3_lifecycle:
    name: mybucket
    prefix: backups/
    storage_class: standard_ia
    transition_days: 31
    state: present
    status: enabled

- name: Configure a lifecycle rule to transition files to infrequent access after 30 days and glacier after 90
  community.aws.s3_lifecycle:
    name: mybucket
    prefix: logs/
    state: present
    status: enabled
    transitions:
      - transition_days: 30
        storage_class: standard_ia
      - transition_days: 90
        storage_class: glacier
"""

import datetime
import time
from copy import deepcopy

try:
    from dateutil import parser as date_parser

    HAS_DATEUTIL = True
except ImportError:
    HAS_DATEUTIL = False

try:
    import botocore
except ImportError:
    pass  # handled by AnsibleAwsModule

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_message
from ansible_collections.amazon.aws.plugins.module_utils.botocore import normalize_boto3_result
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def parse_date(date):
    if date is None:
        return None
    try:
        if HAS_DATEUTIL:
            return date_parser.parse(date)
        else:
            # Very simplistic
            return datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.000Z")
    except ValueError:
        return None


def fetch_rules(client, module, name):
    # Get the bucket's current lifecycle rules
    try:
        current_lifecycle = client.get_bucket_lifecycle_configuration(aws_retry=True, Bucket=name)
        current_lifecycle_rules = normalize_boto3_result(current_lifecycle["Rules"])
    except is_boto3_error_code("NoSuchLifecycleConfiguration"):
        current_lifecycle_rules = []
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e)
    return current_lifecycle_rules


def build_rule(client, module):
    name = module.params.get("name")
    abort_incomplete_multipart_upload_days = module.params.get("abort_incomplete_multipart_upload_days")
    expiration_date = parse_date(module.params.get("expiration_date"))
    expiration_days = module.params.get("expiration_days")
    expire_object_delete_marker = module.params.get("expire_object_delete_marker")
    noncurrent_version_expiration_days = module.params.get("noncurrent_version_expiration_days")
    noncurrent_version_transition_days = module.params.get("noncurrent_version_transition_days")
    noncurrent_version_transitions = module.params.get("noncurrent_version_transitions")
    noncurrent_version_storage_class = module.params.get("noncurrent_version_storage_class")
    noncurrent_version_keep_newer = module.params.get("noncurrent_version_keep_newer")
    prefix = module.params.get("prefix") or ""
    rule_id = module.params.get("rule_id")
    status = module.params.get("status")
    storage_class = module.params.get("storage_class")
    transition_date = parse_date(module.params.get("transition_date"))
    transition_days = module.params.get("transition_days")
    transitions = module.params.get("transitions")
    purge_transitions = module.params.get("purge_transitions")

    rule = dict(Filter=dict(Prefix=prefix), Status=status.title())
    if rule_id is not None:
        rule["ID"] = rule_id

    if abort_incomplete_multipart_upload_days:
        rule["AbortIncompleteMultipartUpload"] = {"DaysAfterInitiation": abort_incomplete_multipart_upload_days}

    # Create expiration
    if expiration_days is not None:
        rule["Expiration"] = dict(Days=expiration_days)
    elif expiration_date is not None:
        rule["Expiration"] = dict(Date=expiration_date.isoformat())
    elif expire_object_delete_marker is not None:
        rule["Expiration"] = dict(ExpiredObjectDeleteMarker=expire_object_delete_marker)
    if noncurrent_version_expiration_days or noncurrent_version_keep_newer:
        rule["NoncurrentVersionExpiration"] = dict()
    if noncurrent_version_expiration_days is not None:
        rule["NoncurrentVersionExpiration"]["NoncurrentDays"] = noncurrent_version_expiration_days
    if noncurrent_version_keep_newer is not None:
        rule["NoncurrentVersionExpiration"]["NewerNoncurrentVersions"] = noncurrent_version_keep_newer
    if transition_days is not None:
        rule["Transitions"] = [
            dict(Days=transition_days, StorageClass=storage_class.upper()),
        ]

    elif transition_date is not None:
        rule["Transitions"] = [
            dict(Date=transition_date.isoformat(), StorageClass=storage_class.upper()),
        ]

    if transitions is not None:
        if not rule.get("Transitions"):
            rule["Transitions"] = []
        for transition in transitions:
            t_out = dict()
            if transition.get("transition_date"):
                t_out["Date"] = transition["transition_date"]
            elif transition.get("transition_days") is not None:
                t_out["Days"] = int(transition["transition_days"])
            if transition.get("storage_class"):
                t_out["StorageClass"] = transition["storage_class"].upper()
                rule["Transitions"].append(t_out)

    if noncurrent_version_transition_days is not None:
        rule["NoncurrentVersionTransitions"] = [
            dict(
                NoncurrentDays=noncurrent_version_transition_days, StorageClass=noncurrent_version_storage_class.upper()
            ),
        ]

    if noncurrent_version_transitions is not None:
        if not rule.get("NoncurrentVersionTransitions"):
            rule["NoncurrentVersionTransitions"] = []
        for noncurrent_version_transition in noncurrent_version_transitions:
            t_out = dict()
            t_out["NoncurrentDays"] = noncurrent_version_transition["transition_days"]
            if noncurrent_version_transition.get("storage_class"):
                t_out["StorageClass"] = noncurrent_version_transition["storage_class"].upper()
                rule["NoncurrentVersionTransitions"].append(t_out)

    return rule


def compare_and_update_configuration(client, module, current_lifecycle_rules, rule):
    purge_transitions = module.params.get("purge_transitions")
    rule_id = module.params.get("rule_id")

    lifecycle_configuration = dict(Rules=[])
    changed = False
    appended = False

    # If current_lifecycle_obj is not None then we have rules to compare, otherwise just add the rule
    if current_lifecycle_rules:
        # If rule ID exists, use that for comparison otherwise compare based on prefix
        for existing_rule in current_lifecycle_rules:
            if rule.get("ID") == existing_rule.get("ID") and rule["Filter"].get("Prefix", "") != existing_rule.get(
                "Filter", {}
            ).get("Prefix", ""):
                existing_rule.pop("ID")
            elif rule_id is None and rule["Filter"].get("Prefix", "") == existing_rule.get("Filter", {}).get(
                "Prefix", ""
            ):
                existing_rule.pop("ID")
            if rule.get("ID") == existing_rule.get("ID"):
                changed_, appended_ = update_or_append_rule(
                    rule, existing_rule, purge_transitions, lifecycle_configuration
                )
                changed = changed_ or changed
                appended = appended_ or appended
            else:
                lifecycle_configuration["Rules"].append(existing_rule)

        # If nothing appended then append now as the rule must not exist
        if not appended:
            lifecycle_configuration["Rules"].append(rule)
            changed = True
    else:
        lifecycle_configuration["Rules"].append(rule)
        changed = True

    return changed, lifecycle_configuration


def update_or_append_rule(new_rule, existing_rule, purge_transitions, lifecycle_obj):
    changed = False
    if existing_rule["Status"] != new_rule["Status"]:
        if not new_rule.get("Transitions") and existing_rule.get("Transitions"):
            new_rule["Transitions"] = existing_rule["Transitions"]
        if not new_rule.get("Expiration") and existing_rule.get("Expiration"):
            new_rule["Expiration"] = existing_rule["Expiration"]
        if not new_rule.get("NoncurrentVersionExpiration") and existing_rule.get("NoncurrentVersionExpiration"):
            new_rule["NoncurrentVersionExpiration"] = existing_rule["NoncurrentVersionExpiration"]
        lifecycle_obj["Rules"].append(new_rule)
        changed = True
        appended = True
    else:
        if not purge_transitions:
            merge_transitions(new_rule, existing_rule)
        if compare_rule(new_rule, existing_rule, purge_transitions):
            lifecycle_obj["Rules"].append(new_rule)
            appended = True
        else:
            lifecycle_obj["Rules"].append(new_rule)
            changed = True
            appended = True
    return changed, appended


def compare_and_remove_rule(current_lifecycle_rules, rule_id=None, prefix=None):
    changed = False
    lifecycle_configuration = dict(Rules=[])

    # Check if rule exists
    # If an ID exists, use that otherwise compare based on prefix
    if rule_id is not None:
        for existing_rule in current_lifecycle_rules:
            if rule_id == existing_rule["ID"]:
                # We're not keeping the rule (i.e. deleting) so mark as changed
                changed = True
            else:
                lifecycle_configuration["Rules"].append(existing_rule)
    else:
        for existing_rule in current_lifecycle_rules:
            if prefix == existing_rule["Filter"].get("Prefix", ""):
                # We're not keeping the rule (i.e. deleting) so mark as changed
                changed = True
            else:
                lifecycle_configuration["Rules"].append(existing_rule)

    return changed, lifecycle_configuration


def compare_rule(new_rule, old_rule, purge_transitions):
    # Copy objects
    rule1 = deepcopy(new_rule)
    rule2 = deepcopy(old_rule)

    if purge_transitions:
        return rule1 == rule2
    else:
        transitions1 = rule1.pop("Transitions", [])
        transitions2 = rule2.pop("Transitions", [])
        noncurrent_transtions1 = rule1.pop("NoncurrentVersionTransitions", [])
        noncurrent_transtions2 = rule2.pop("NoncurrentVersionTransitions", [])
        if rule1 != rule2:
            return False
        for transition in transitions1:
            if transition not in transitions2:
                return False
        for transition in noncurrent_transtions1:
            if transition not in noncurrent_transtions2:
                return False
        return True


def merge_transitions(updated_rule, updating_rule):
    # because of the legal S3 transitions, we know only one can exist for each storage class.
    # So, our strategy is build some dicts, keyed on storage class and add the storage class transitions that are only
    # in updating_rule to updated_rule
    updated_transitions = {}
    updating_transitions = {}
    for transition in updated_rule.get("Transitions", []):
        updated_transitions[transition["StorageClass"]] = transition
    for transition in updating_rule.get("Transitions", []):
        updating_transitions[transition["StorageClass"]] = transition
    for storage_class, transition in updating_transitions.items():
        if updated_transitions.get(storage_class) is None:
            updated_rule["Transitions"].append(transition)


def create_lifecycle_rule(client, module):
    name = module.params.get("name")
    wait = module.params.get("wait")
    changed = False

    old_lifecycle_rules = fetch_rules(client, module, name)
    new_rule = build_rule(client, module)
    (changed, lifecycle_configuration) = compare_and_update_configuration(client, module, old_lifecycle_rules, new_rule)
    if changed:
        # Write lifecycle to bucket
        try:
            client.put_bucket_lifecycle_configuration(
                aws_retry=True, Bucket=name, LifecycleConfiguration=lifecycle_configuration
            )
        except is_boto3_error_message("At least one action needs to be specified in a rule"):
            # Amazon interpreted this as not changing anything
            changed = False
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
        ) as e:  # pylint: disable=duplicate-except
            module.fail_json_aws(
                e, lifecycle_configuration=lifecycle_configuration, name=name, old_lifecycle_rules=old_lifecycle_rules
            )

        _changed = changed
        _retries = 10
        _not_changed_cnt = 6
        while wait and _changed and _retries and _not_changed_cnt:
            # We've seen examples where get_bucket_lifecycle_configuration returns
            # the updated rules, then the old rules, then the updated rules again and
            # again couple of times.
            # Thus try to read the rule few times in a row to check if it has changed.
            time.sleep(5)
            _retries -= 1
            new_rules = fetch_rules(client, module, name)
            (_changed, lifecycle_configuration) = compare_and_update_configuration(client, module, new_rules, new_rule)
            if not _changed:
                _not_changed_cnt -= 1
                _changed = True
            else:
                _not_changed_cnt = 6
    else:
        _retries = 0

    new_rules = fetch_rules(client, module, name)

    module.exit_json(
        changed=changed,
        new_rule=new_rule,
        rules=new_rules,
        old_rules=old_lifecycle_rules,
        _retries=_retries,
        _config=lifecycle_configuration,
    )


def destroy_lifecycle_rule(client, module):
    name = module.params.get("name")
    prefix = module.params.get("prefix")
    rule_id = module.params.get("rule_id")
    wait = module.params.get("wait")
    changed = False

    if prefix is None:
        prefix = ""

    current_lifecycle_rules = fetch_rules(client, module, name)
    changed, lifecycle_obj = compare_and_remove_rule(current_lifecycle_rules, rule_id, prefix)

    if changed:
        # Write lifecycle to bucket or, if there no rules left, delete lifecycle configuration
        try:
            if lifecycle_obj["Rules"]:
                client.put_bucket_lifecycle_configuration(
                    aws_retry=True, Bucket=name, LifecycleConfiguration=lifecycle_obj
                )
            elif current_lifecycle_rules:
                changed = True
                client.delete_bucket_lifecycle(aws_retry=True, Bucket=name)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e)

        _changed = changed
        _retries = 10
        _not_changed_cnt = 6
        while wait and _changed and _retries and _not_changed_cnt:
            # We've seen examples where get_bucket_lifecycle_configuration returns
            # the updated rules, then the old rules, then the updated rules again and
            # again couple of times.
            # Thus try to read the rule few times in a row to check if it has changed.
            time.sleep(5)
            _retries -= 1
            new_rules = fetch_rules(client, module, name)
            (_changed, lifecycle_configuration) = compare_and_remove_rule(new_rules, rule_id, prefix)
            if not _changed:
                _not_changed_cnt -= 1
                _changed = True
            else:
                _not_changed_cnt = 6
    else:
        _retries = 0

    new_rules = fetch_rules(client, module, name)

    module.exit_json(changed=changed, rules=new_rules, old_rules=current_lifecycle_rules, _retries=_retries)


def main():
    s3_storage_class = ["glacier", "onezone_ia", "standard_ia", "intelligent_tiering", "deep_archive"]
    argument_spec = dict(
        name=dict(required=True, type="str"),
        abort_incomplete_multipart_upload_days=dict(type="int"),
        expiration_days=dict(type="int"),
        expiration_date=dict(),
        expire_object_delete_marker=dict(type="bool"),
        noncurrent_version_expiration_days=dict(type="int"),
        noncurrent_version_keep_newer=dict(type="int"),
        noncurrent_version_storage_class=dict(default="glacier", type="str", choices=s3_storage_class),
        noncurrent_version_transition_days=dict(type="int"),
        noncurrent_version_transitions=dict(type="list", elements="dict"),
        prefix=dict(),
        rule_id=dict(),
        state=dict(default="present", choices=["present", "absent"]),
        status=dict(default="enabled", choices=["enabled", "disabled"]),
        storage_class=dict(default="glacier", type="str", choices=s3_storage_class),
        transition_days=dict(type="int"),
        transition_date=dict(),
        transitions=dict(type="list", elements="dict"),
        purge_transitions=dict(default=True, type="bool"),
        wait=dict(type="bool", default=False),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["expiration_days", "expiration_date", "expire_object_delete_marker"],
            ["expiration_days", "transition_date"],
            ["transition_days", "transition_date"],
            ["transition_days", "expiration_date"],
            ["transition_days", "transitions"],
            ["transition_date", "transitions"],
            ["noncurrent_version_transition_days", "noncurrent_version_transitions"],
        ],
        required_by={
            "noncurrent_version_keep_newer": ["noncurrent_version_expiration_days"],
        },
    )

    client = module.client("s3", retry_decorator=AWSRetry.jittered_backoff())

    expiration_date = module.params.get("expiration_date")
    transition_date = module.params.get("transition_date")
    state = module.params.get("state")

    if state == "present" and module.params["status"] == "enabled":  # allow deleting/disabling a rule by id/prefix
        required_when_present = (
            "abort_incomplete_multipart_upload_days",
            "expiration_date",
            "expiration_days",
            "expire_object_delete_marker",
            "transition_date",
            "transition_days",
            "transitions",
            "noncurrent_version_expiration_days",
            "noncurrent_version_keep_newer",
            "noncurrent_version_transition_days",
            "noncurrent_version_transitions",
        )
        for param in required_when_present:
            if module.params.get(param) is None:
                break
        else:
            msg = f"one of the following is required when 'state' is 'present': {', '.join(required_when_present)}"
            module.fail_json(msg=msg)

    # If dates have been set, make sure they're in a valid format
    if expiration_date:
        expiration_date = parse_date(expiration_date)
        if expiration_date is None:
            module.fail_json(
                msg="expiration_date is not a valid ISO-8601 format."
                "  The time must be midnight and a timezone of GMT must be included"
            )
    if transition_date:
        transition_date = parse_date(transition_date)
        if transition_date is None:
            module.fail_json(
                msg="transition_date is not a valid ISO-8601 format."
                "  The time must be midnight and a timezone of GMT must be included"
            )

    if state == "present":
        create_lifecycle_rule(client, module)
    elif state == "absent":
        destroy_lifecycle_rule(client, module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020-2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type


import abc
import traceback

from ansible.module_utils import six
from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.community.crypto.plugins.module_utils.version import LooseVersion

from ansible_collections.community.crypto.plugins.module_utils.crypto.basic import (
    CRYPTOGRAPHY_HAS_X25519,
    CRYPTOGRAPHY_HAS_X448,
    CRYPTOGRAPHY_HAS_ED25519,
    CRYPTOGRAPHY_HAS_ED448,
    OpenSSLObjectError,
)

from ansible_collections.community.crypto.plugins.module_utils.crypto.support import (
    get_fingerprint_of_bytes,
    load_publickey,
)


MINIMAL_CRYPTOGRAPHY_VERSION = '1.2.3'

CRYPTOGRAPHY_IMP_ERR = None
try:
    import cryptography
    from cryptography.hazmat.primitives import serialization
    CRYPTOGRAPHY_VERSION = LooseVersion(cryptography.__version__)
except ImportError:
    CRYPTOGRAPHY_IMP_ERR = traceback.format_exc()
    CRYPTOGRAPHY_FOUND = False
else:
    CRYPTOGRAPHY_FOUND = True


def _get_cryptography_public_key_info(key):
    key_public_data = dict()
    if isinstance(key, cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey):
        key_type = 'RSA'
        public_numbers = key.public_numbers()
        key_public_data['size'] = key.key_size
        key_public_data['modulus'] = public_numbers.n
        key_public_data['exponent'] = public_numbers.e
    elif isinstance(key, cryptography.hazmat.primitives.asymmetric.dsa.DSAPublicKey):
        key_type = 'DSA'
        parameter_numbers = key.parameters().parameter_numbers()
        public_numbers = key.public_numbers()
        key_public_data['size'] = key.key_size
        key_public_data['p'] = parameter_numbers.p
        key_public_data['q'] = parameter_numbers.q
        key_public_data['g'] = parameter_numbers.g
        key_public_data['y'] = public_numbers.y
    elif CRYPTOGRAPHY_HAS_X25519 and isinstance(key, cryptography.hazmat.primitives.asymmetric.x25519.X25519PublicKey):
        key_type = 'X25519'
    elif CRYPTOGRAPHY_HAS_X448 and isinstance(key, cryptography.hazmat.primitives.asymmetric.x448.X448PublicKey):
        key_type = 'X448'
    elif CRYPTOGRAPHY_HAS_ED25519 and isinstance(key, cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey):
        key_type = 'Ed25519'
    elif CRYPTOGRAPHY_HAS_ED448 and isinstance(key, cryptography.hazmat.primitives.asymmetric.ed448.Ed448PublicKey):
        key_type = 'Ed448'
    elif isinstance(key, cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey):
        key_type = 'ECC'
        public_numbers = key.public_numbers()
        key_public_data['curve'] = key.curve.name
        key_public_data['x'] = public_numbers.x
        key_public_data['y'] = public_numbers.y
        key_public_data['exponent_size'] = key.curve.key_size
    else:
        key_type = 'unknown ({0})'.format(type(key))
    return key_type, key_public_data


class PublicKeyParseError(OpenSSLObjectError):
    def __init__(self, msg, result):
        super(PublicKeyParseError, self).__init__(msg)
        self.error_message = msg
        self.result = result


@six.add_metaclass(abc.ABCMeta)
class PublicKeyInfoRetrieval(object):
    def __init__(self, module, backend, content=None, key=None):
        # content must be a bytes string
        self.module = module
        self.backend = backend
        self.content = content
        self.key = key

    @abc.abstractmethod
    def _get_public_key(self, binary):
        pass

    @abc.abstractmethod
    def _get_key_info(self):
        pass

    def get_info(self, prefer_one_fingerprint=False):
        result = dict()
        if self.key is None:
            try:
                self.key = load_publickey(content=self.content, backend=self.backend)
            except OpenSSLObjectError as e:
                raise PublicKeyParseError(to_native(e), {})

        pk = self._get_public_key(binary=True)
        result['fingerprints'] = get_fingerprint_of_bytes(
            pk, prefer_one=prefer_one_fingerprint) if pk is not None else dict()

        key_type, key_public_data = self._get_key_info()
        result['type'] = key_type
        result['public_data'] = key_public_data
        return result


class PublicKeyInfoRetrievalCryptography(PublicKeyInfoRetrieval):
    """Validate the supplied public key, using the cryptography backend"""
    def __init__(self, module, content=None, key=None):
        super(PublicKeyInfoRetrievalCryptography, self).__init__(module, 'cryptography', content=content, key=key)

    def _get_public_key(self, binary):
        return self.key.public_bytes(
            serialization.Encoding.DER if binary else serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo
        )

    def _get_key_info(self):
        return _get_cryptography_public_key_info(self.key)


def get_publickey_info(module, backend, content=None, key=None, prefer_one_fingerprint=False):
    if backend == 'cryptography':
        info = PublicKeyInfoRetrievalCryptography(module, content=content, key=key)
    return info.get_info(prefer_one_fingerprint=prefer_one_fingerprint)


def select_backend(module, backend, content=None, key=None):
    if backend == 'auto':
        # Detection what is possible
        can_use_cryptography = CRYPTOGRAPHY_FOUND and CRYPTOGRAPHY_VERSION >= LooseVersion(MINIMAL_CRYPTOGRAPHY_VERSION)

        # Try cryptography
        if can_use_cryptography:
            backend = 'cryptography'

        # Success?
        if backend == 'auto':
            module.fail_json(msg=("Cannot detect any of the required Python libraries "
                                  "cryptography (>= {0})").format(MINIMAL_CRYPTOGRAPHY_VERSION))

    if backend == 'cryptography':
        if not CRYPTOGRAPHY_FOUND:
            module.fail_json(msg=missing_required_lib('cryptography >= {0}'.format(MINIMAL_CRYPTOGRAPHY_VERSION)),
                             exception=CRYPTOGRAPHY_IMP_ERR)
        return backend, PublicKeyInfoRetrievalCryptography(module, content=content, key=key)
    else:
        raise ValueError('Unsupported value for backend: {0}'.format(backend))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: site_update
short_description: Resource module for Site Update
description:
- Manage operation update of the resource Site Update.
- Update site area/building/floor with specified hierarchy and new values.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  site:
    description: Site Update's site.
    suboptions:
      area:
        description: Site Update's area.
        suboptions:
          name:
            description: Name.
            type: str
          parentName:
            description: Parent Name.
            type: str
        type: dict
      building:
        description: Site Update's building.
        suboptions:
          address:
            description: Address.
            type: str
          latitude:
            description: Latitude.
            type: int
          longitude:
            description: Longitude.
            type: int
          name:
            description: Name.
            type: str
          parentName:
            description: Parent Name.
            type: str
        type: dict
      floor:
        description: Site Update's floor.
        suboptions:
          height:
            description: Height.
            type: int
          length:
            description: Length.
            type: int
          name:
            description: Name.
            type: str
          rfModel:
            description: Rf Model.
            type: str
          width:
            description: Width.
            type: int
        type: dict
    type: dict
  siteId:
    description: SiteId path parameter. Site id to which site details to be updated.
    type: str
  type:
    description: Type.
    type: str
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Sites UpdateSite
  description: Complete reference of the UpdateSite API.
  link: https://developer.cisco.com/docs/dna-center/#!update-site
notes:
  - SDK Method used are
    sites.Sites.update_site,

  - Paths used are
    put /dna/intent/api/v1/site/{siteId},

"""

EXAMPLES = r"""
- name: Update by id
  cisco.dnac.site_update:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: '{{my_headers | from_json}}'
    site:
      area:
        name: string
        parentName: string
      building:
        address: string
        latitude: 0
        longitude: 0
        name: string
        parentName: string
      floor:
        height: 0
        length: 0
        name: string
        rfModel: string
        width: 0
    siteId: string
    type: string

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "result": "string",
      "response": {
        "endTime": "string",
        "version": "string",
        "startTime": "string",
        "progress": "string",
        "data": "string",
        "serviceType": "string",
        "operationIdList": [
          "string"
        ],
        "isError": "string",
        "rootId": "string",
        "instanceTenantId": "string",
        "id": "string"
      },
      "status": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Yaakov Kuperman <ykuperman@gmail.com>
# GNU General Public License v3.0+ # (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: elb_target_info
version_added: 1.0.0
short_description: Gathers which target groups a target is associated with.
description:
  - This module will search through every target group in a region to find
    which ones have registered a given instance ID or IP.
author:
  - "Yaakov Kuperman (@yaakov-github)"
options:
  instance_id:
    description:
      - What instance ID to get information for.
    type: str
    required: true
  get_unused_target_groups:
    description:
      - Whether or not to get target groups not used by any load balancers.
    type: bool
    default: true

extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# practical use case - dynamically de-registering and re-registering nodes

- name: Get EC2 Metadata
  amazon.aws.ec2_metadata_facts:

- name: Get initial list of target groups
  delegate_to: localhost
  community.aws.elb_target_info:
    instance_id: "{{ ansible_ec2_instance_id }}"
    region: "{{ ansible_ec2_placement_region }}"
  register: target_info

- name: save fact for later
  ansible.builtin.set_fact:
    original_tgs: "{{ target_info.instance_target_groups }}"

- name: Deregister instance from all target groups
  delegate_to: localhost
  community.aws.elb_target:
    target_group_arn: "{{ item.0.target_group_arn }}"
    target_port: "{{ item.1.target_port }}"
    target_az: "{{ item.1.target_az }}"
    target_id: "{{ item.1.target_id }}"
    state: absent
    target_status: "draining"
    region: "{{ ansible_ec2_placement_region }}"
  with_subelements:
    - "{{ original_tgs }}"
    - "targets"

  # This avoids having to wait for 'elb_target' to serially deregister each
  # target group.  An alternative would be to run all of the 'elb_target'
  # tasks async and wait for them to finish.

- name: wait for all targets to deregister simultaneously
  delegate_to: localhost
  community.aws.elb_target_info:
    get_unused_target_groups: false
    instance_id: "{{ ansible_ec2_instance_id }}"
    region: "{{ ansible_ec2_placement_region }}"
  register: target_info
  until: (target_info.instance_target_groups | length) == 0
  retries: 60
  delay: 10

- name: reregister in elbv2s
  community.aws.elb_target:
    region: "{{ ansible_ec2_placement_region }}"
    target_group_arn: "{{ item.0.target_group_arn }}"
    target_port: "{{ item.1.target_port }}"
    target_az: "{{ item.1.target_az }}"
    target_id: "{{ item.1.target_id }}"
    state: present
    target_status: "initial"
  with_subelements:
    - "{{ original_tgs }}"
    - "targets"

# wait until all groups associated with this instance are 'healthy' or
# 'unused'
- name: wait for registration
  community.aws.elb_target_info:
    get_unused_target_groups: false
    instance_id: "{{ ansible_ec2_instance_id }}"
    region: "{{ ansible_ec2_placement_region }}"
  register: target_info
  until: (target_info.instance_target_groups |
          map(attribute='targets') |
          flatten |
          map(attribute='target_health') |
          rejectattr('state', 'equalto', 'healthy') |
          rejectattr('state', 'equalto', 'unused') |
          list |
          length) == 0
  retries: 61
  delay: 10

# using the target groups to generate AWS CLI commands to reregister the
# instance - useful in case the playbook fails mid-run and manual
#            rollback is required
- name: "reregistration commands: ELBv2s"
  ansible.builtin.debug:
    msg: >
           aws --region {{ansible_ec2_placement_region}} elbv2
           register-targets --target-group-arn {{item.target_group_arn}}
           --targets{%for target in item.targets%}
           Id={{target.target_id}},
           Port={{target.target_port}}{%if target.target_az%},AvailabilityZone={{target.target_az}}
           {%endif%}
           {%endfor%}
  loop: "{{target_info.instance_target_groups}}"
"""

RETURN = r"""
instance_target_groups:
    description: a list of target groups to which the instance is registered to
    returned: always
    type: complex
    contains:
        target_group_arn:
            description: The ARN of the target group
            type: str
            returned: always
            sample:
                - "arn:aws:elasticloadbalancing:eu-west-1:123456789012:targetgroup/target-group/deadbeefdeadbeef"
        target_group_type:
            description: Which target type is used for this group
            returned: always
            type: str
            sample:
                - ip
                - instance
        targets:
            description: A list of targets that point to this instance ID
            returned: always
            type: complex
            contains:
                target_id:
                    description: the target ID referring to this instance
                    type: str
                    returned: always
                    sample:
                        - i-deadbeef
                        - 1.2.3.4
                target_port:
                    description: which port this target is listening on
                    type: str
                    returned: always
                    sample:
                        - 80
                target_az:
                    description: which availability zone is explicitly
                                 associated with this target
                    type: str
                    returned: when an AZ is associated with this instance
                    sample:
                        - us-west-2a
                target_health:
                    description:
                    - The target health description.
                    - See following link for all the possible values
                      U(https://boto3.readthedocs.io/en/latest/reference/services/elbv2.html#ElasticLoadBalancingv2.Client.describe_target_health)
                    returned: always
                    type: complex
                    contains:
                        description:
                            description: description of target health
                            returned: if I(state!=present)
                            sample:
                                - "Target desregistration is in progress"
                            type: str
                        reason:
                            description: reason code for target health
                            returned: if I(state!=healthy)
                            sample:
                                - "Target.Deregistration in progress"
                            type: str
                        state:
                            description: health state
                            returned: always
                            sample:
                                - "healthy"
                                - "draining"
                                - "initial"
                                - "unhealthy"
                                - "unused"
                                - "unavailable"
                            type: str
"""

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    # we can handle the lack of boto3 based on the ec2 module
    pass

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


class Target(object):
    """Models a target in a target group"""

    def __init__(self, target_id, port, az, raw_target_health):
        self.target_port = port
        self.target_id = target_id
        self.target_az = az
        self.target_health = self.convert_target_health(raw_target_health)

    def convert_target_health(self, raw_target_health):
        return camel_dict_to_snake_dict(raw_target_health)


class TargetGroup(object):
    """Models an elbv2 target group"""

    def __init__(self, **kwargs):
        self.target_group_type = kwargs["target_group_type"]
        self.target_group_arn = kwargs["target_group_arn"]
        # the relevant targets associated with this group
        self.targets = []

    def add_target(self, target_id, target_port, target_az, raw_target_health):
        self.targets.append(Target(target_id, target_port, target_az, raw_target_health))

    def to_dict(self):
        object_dict = vars(self)
        object_dict["targets"] = [vars(each) for each in self.get_targets()]
        return object_dict

    def get_targets(self):
        return list(self.targets)


class TargetInfoGatherer(object):
    def __init__(self, module, instance_id, get_unused_target_groups):
        self.module = module
        try:
            self.ec2 = self.module.client("ec2", retry_decorator=AWSRetry.jittered_backoff(retries=10))
        except (ClientError, BotoCoreError) as e:
            self.module.fail_json_aws(e, msg="Couldn't connect to ec2")

        try:
            self.elbv2 = self.module.client("elbv2", retry_decorator=AWSRetry.jittered_backoff(retries=10))
        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(e, msg="Could not connect to elbv2")

        self.instance_id = instance_id
        self.get_unused_target_groups = get_unused_target_groups
        self.tgs = self._get_target_groups()

    def _get_instance_ips(self):
        """Fetch all IPs associated with this instance so that we can determine
        whether or not an instance is in an IP-based target group"""
        try:
            # get ahold of the instance in the API
            reservations = self.ec2.describe_instances(InstanceIds=[self.instance_id], aws_retry=True)["Reservations"]
        except (BotoCoreError, ClientError) as e:
            # typically this will happen if the instance doesn't exist
            self.module.fail_json_aws(
                e,
                msg=f"Could not get instance info for instance '{self.instance_id}'",
            )

        if len(reservations) < 1:
            self.module.fail_json(msg=f"Instance ID {self.instance_id} could not be found")

        instance = reservations[0]["Instances"][0]

        # IPs are represented in a few places in the API, this should
        # account for all of them
        ips = set()
        ips.add(instance["PrivateIpAddress"])
        for nic in instance["NetworkInterfaces"]:
            ips.add(nic["PrivateIpAddress"])
            for ip in nic["PrivateIpAddresses"]:
                ips.add(ip["PrivateIpAddress"])

        return list(ips)

    def _get_target_group_objects(self):
        """helper function to build a list of TargetGroup objects based on
        the AWS API"""
        try:
            paginator = self.elbv2.get_paginator("describe_target_groups")
            tg_response = paginator.paginate().build_full_result()
        except (BotoCoreError, ClientError) as e:
            self.module.fail_json_aws(
                e,
                msg="Could not describe target groups",
            )

        # build list of TargetGroup objects representing every target group in
        # the system
        target_groups = []
        for each_tg in tg_response["TargetGroups"]:
            if not self.get_unused_target_groups and len(each_tg["LoadBalancerArns"]) < 1:
                # only collect target groups that actually are connected
                # to LBs
                continue

            target_groups.append(
                TargetGroup(
                    target_group_arn=each_tg["TargetGroupArn"],
                    target_group_type=each_tg["TargetType"],
                )
            )
        return target_groups

    def _get_target_descriptions(self, target_groups):
        """Helper function to build a list of all the target descriptions
        for this target in a target group"""
        # Build a list of all the target groups pointing to this instance
        # based on the previous list
        tgs = set()
        # Loop through all the target groups
        for tg in target_groups:
            try:
                # Get the list of targets for that target group
                response = self.elbv2.describe_target_health(TargetGroupArn=tg.target_group_arn, aws_retry=True)
            except (BotoCoreError, ClientError) as e:
                self.module.fail_json_aws(
                    e, msg="Could not describe target " + f"health for target group {tg.target_group_arn}"
                )

            for t in response["TargetHealthDescriptions"]:
                # If the target group has this instance as a target, add to
                # list. This logic also accounts for the possibility of a
                # target being in the target group multiple times with
                # overridden ports
                if t["Target"]["Id"] == self.instance_id or t["Target"]["Id"] in self.instance_ips:
                    # The 'AvailabilityZone' parameter is a weird one, see the
                    # API docs for more.  Basically it's only supposed to be
                    # there under very specific circumstances, so we need
                    # to account for that
                    az = t["Target"]["AvailabilityZone"] if "AvailabilityZone" in t["Target"] else None

                    tg.add_target(t["Target"]["Id"], t["Target"]["Port"], az, t["TargetHealth"])
                    # since tgs is a set, each target group will be added only
                    # once, even though we call add on each successful match
                    tgs.add(tg)
        return list(tgs)

    def _get_target_groups(self):
        # do this first since we need the IPs later on in this function
        self.instance_ips = self._get_instance_ips()

        # build list of target groups
        target_groups = self._get_target_group_objects()
        return self._get_target_descriptions(target_groups)


def main():
    argument_spec = dict(
        instance_id={"required": True, "type": "str"},
        get_unused_target_groups={"required": False, "default": True, "type": "bool"},
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    instance_id = module.params["instance_id"]
    get_unused_target_groups = module.params["get_unused_target_groups"]

    tg_gatherer = TargetInfoGatherer(module, instance_id, get_unused_target_groups)

    instance_target_groups = [each.to_dict() for each in tg_gatherer.tgs]

    module.exit_json(instance_target_groups=instance_target_groups)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: pnp_virtual_account_deregister
short_description: Resource module for Pnp Virtual Account Deregister
description:
- Manage operation delete of the resource Pnp Virtual Account Deregister.
- >
   Deregisters the specified smart account & virtual account info and the associated device information from the PnP
   System & database. The devices associated with the deregistered virtual account are removed from the PnP database
   as well. The response payload contains the deregistered smart & virtual account information.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  domain:
    description: Domain query parameter. Smart Account Domain.
    type: str
  name:
    description: Name query parameter. Virtual Account Name.
    type: str
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Device Onboarding (PnP) DeregisterVirtualAccount
  description: Complete reference of the DeregisterVirtualAccount API.
  link: https://developer.cisco.com/docs/dna-center/#!deregister-virtual-account
notes:
  - SDK Method used are
    device_onboarding_pnp.DeviceOnboardingPnp.deregister_virtual_account,

  - Paths used are
    delete /dna/intent/api/v1/onboarding/pnp-settings/vacct,

"""

EXAMPLES = r"""
- name: Delete all
  cisco.dnac.pnp_virtual_account_deregister:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    domain: string
    name: string

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "virtualAccountId": "string",
      "autoSyncPeriod": 0,
      "syncResultStr": "string",
      "profile": {
        "proxy": true,
        "makeDefault": true,
        "port": 0,
        "profileId": "string",
        "name": "string",
        "addressIpV4": "string",
        "cert": "string",
        "addressFqdn": "string"
      },
      "ccoUser": "string",
      "syncResult": {
        "syncList": [
          {
            "syncType": "string",
            "deviceSnList": [
              "string"
            ]
          }
        ],
        "syncMsg": "string"
      },
      "token": "string",
      "syncStartTime": 0,
      "lastSync": 0,
      "tenantId": "string",
      "smartAccountId": "string",
      "expiry": 0,
      "syncStatus": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: node_promotion
short_description: Resource module for Node Promotion
description:
- Manage operation update of the resource Node Promotion.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.ise.module
author: Rafael Campos (@racampos)
options:
  promotionType:
    description: Node Promotion's promotionType.
    type: str
requirements:
- ciscoisesdk >= 2.1.1
- python >= 3.5
notes:
  - SDK Method used are
    node_deployment.NodeDeployment.promote_node,

  - Paths used are
    put /api/v1/deployment/node-promotion/,
"""

EXAMPLES = r"""
- name: Update all
  cisco.ise.node_promotion:
    ise_hostname: "{{ise_hostname}}"
    ise_username: "{{ise_username}}"
    ise_password: "{{ise_password}}"
    ise_verify: "{{ise_verify}}"
    promotionType: string

"""

RETURN = r"""
ise_response:
  description: A dictionary or list with the response returned by the Cisco ISE Python SDK
  returned: always
  type: dict
  sample: >
    {
      "code": 0,
      "message": "string",
      "rootCause": "string"
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2016 Matt Davis, <mdavis@ansible.com>
#                    Chris Houseknecht, <house@redhat.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_securitygroup_info

version_added: "0.1.2"

short_description: Get security group facts

description:
    - Get facts for a specific security group or all security groups within a resource group.

options:
    name:
        description:
            - Only show results for a specific security group.
        type: str
    resource_group:
        description:
            - Name of the resource group to use.
        required: true
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Chris Houseknecht (@chouseknecht)
    - Matt Davis (@nitzmahone)

'''

EXAMPLES = '''
- name: Get facts for one security group
  azure_rm_securitygroup_info:
    resource_group: myResourceGroup
    name: secgroup001

- name: Get facts for all security groups
  azure_rm_securitygroup_info:
    resource_group: myResourceGroup
'''

RETURN = '''
securitygroups:
    description:
        - List containing security group dicts.
    returned: always
    type: complex
    contains:
        etag:
            description:
                - A unique read-only string that changes whenever the resource is updated.
            returned: always
            type: str
            sample:  'W/"d036f4d7-d977-429a-a8c6-879bc2523399"'
        id:
            description:
                - Resource ID.
            returned: always
            type: str
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/secgroup001"
        location:
            description:
                - Resource location.
            returned: always
            type: str
            sample: "eastus2"
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: "secgroup001"
        default_rules:
            description:
                - The default security rules of network security group.
            returned: always
            type: list
            sample: [
                    {
                        "access": "Allow",
                        "description": "Allow inbound traffic from all VMs in VNET",
                        "destination_address_prefix": "VirtualNetwork",
                        "destination_port_range": "*",
                        "direction": "Inbound",
                        "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/defaultSecurityRules/AllowVnetInBound",
                        "name": "AllowVnetInBound",
                        "priority": 65000,
                        "protocol": "*",
                        "provisioning_state": "Succeeded",
                        "source_address_prefix": "VirtualNetwork",
                        "source_port_range": "*"
                    },
                    {
                        "access": "Allow",
                        "description": "Allow inbound traffic from azure load balancer",
                        "destination_address_prefix": "*",
                        "destination_port_range": "*",
                        "direction": "Inbound",
                        "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/defaultSecurityRules/AllowAzureLoadBalancerInBound",
                        "name": "AllowAzureLoadBalancerInBound",
                        "priority": 65001,
                        "protocol": "*",
                        "provisioning_state": "Succeeded",
                        "source_address_prefix": "AzureLoadBalancer",
                        "source_port_range": "*"
                    },
                    {
                        "access": "Deny",
                        "description": "Deny all inbound traffic",
                        "destination_address_prefix": "*",
                        "destination_port_range": "*",
                        "direction": "Inbound",
                        "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/defaultSecurityRules/DenyAllInBound",
                        "name": "DenyAllInBound",
                        "priority": 65500,
                        "protocol": "*",
                        "provisioning_state": "Succeeded",
                        "source_address_prefix": "*",
                        "source_port_range": "*"
                    },
                    {
                        "access": "Allow",
                        "description": "Allow outbound traffic from all VMs to all VMs in VNET",
                        "destination_address_prefix": "VirtualNetwork",
                        "destination_port_range": "*",
                        "direction": "Outbound",
                        "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/defaultSecurityRules/AllowVnetOutBound",
                        "name": "AllowVnetOutBound",
                        "priority": 65000,
                        "protocol": "*",
                        "provisioning_state": "Succeeded",
                        "source_address_prefix": "VirtualNetwork",
                        "source_port_range": "*"
                    },
                    {
                        "access": "Allow",
                        "description": "Allow outbound traffic from all VMs to Internet",
                        "destination_address_prefix": "Internet",
                        "destination_port_range": "*",
                        "direction": "Outbound",
                        "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/defaultSecurityRules/AllowInternetOutBound",
                        "name": "AllowInternetOutBound",
                        "priority": 65001,
                        "protocol": "*",
                        "provisioning_state": "Succeeded",
                        "source_address_prefix": "*",
                        "source_port_range": "*"
                    },
                    {
                        "access": "Deny",
                        "description": "Deny all outbound traffic",
                        "destination_address_prefix": "*",
                        "destination_port_range": "*",
                        "direction": "Outbound",
                        "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/defaultSecurityRules/DenyAllOutBound",
                        "name": "DenyAllOutBound",
                        "priority": 65500,
                        "protocol": "*",
                        "provisioning_state": "Succeeded",
                        "source_address_prefix": "*",
                        "source_port_range": "*"
                    }
                ]
        network_interfaces:
            description:
                - A collection of references to network interfaces.
            returned: always
            type: list
            sample: []
        rules:
            description:
                - A collection of security rules of the network security group.
            returned: always
            type: list
            sample: [
                {
                    "access": "Deny",
                    "description": null,
                    "destination_address_prefix": "*",
                    "destination_port_range": "22",
                    "direction": "Inbound",
                    "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                    "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/securityRules/DenySSH",
                    "name": "DenySSH",
                    "priority": 100,
                    "protocol": "Tcp",
                    "provisioning_state": "Succeeded",
                    "source_address_prefix": "*",
                    "source_port_range": "*"
                },
                {
                    "access": "Allow",
                    "description": null,
                    "destination_address_prefix": "*",
                    "destination_port_range": "22",
                    "direction": "Inbound",
                    "etag": 'W/"edf48d56-b315-40ca-a85d-dbcb47f2da7d"',
                    "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/mysecgroup/securityRules/AllowSSH",
                    "name": "AllowSSH",
                    "priority": 101,
                    "protocol": "Tcp",
                    "provisioning_state": "Succeeded",
                    "source_address_prefix": "174.109.158.0/24",
                    "source_port_range": "*"
                }
            ]
        subnets:
            description:
                - A collection of references to subnets.
            returned: always
            type: list
            sample: []
        tags:
            description:
                - Tags to assign to the security group.
            returned: always
            type: dict
            sample: { 'tag': 'value' }
        type:
            description:
                - Type of the resource.
            returned: always
            type: str
            sample: "Microsoft.Network/networkSecurityGroups"

'''  # NOQA

try:
    from azure.core.exceptions import ResourceNotFoundError
except Exception:
    # This is handled in azure_rm_common
    pass

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


def create_rule_dict_from_obj(rule):
    return dict(
        id=rule.id,
        name=rule.name,
        description=rule.description,
        protocol=rule.protocol,
        source_port_range=rule.source_port_range,
        destination_port_range=rule.destination_port_range,
        source_address_prefix=rule.source_address_prefix,
        destination_address_prefix=rule.destination_address_prefix,
        source_port_ranges=rule.source_port_ranges,
        destination_port_ranges=rule.destination_port_ranges,
        source_address_prefixes=rule.source_address_prefixes,
        destination_address_prefixes=rule.destination_address_prefixes,
        source_application_security_groups=[p.id for p in rule.source_application_security_groups] if rule.source_application_security_groups else None,
        destination_application_security_groups=[
            p.id for p in rule.destination_application_security_groups] if rule.destination_application_security_groups else None,
        access=rule.access,
        priority=rule.priority,
        direction=rule.direction,
        provisioning_state=rule.provisioning_state,
        etag=rule.etag
    )


def create_network_security_group_dict(nsg):
    result = dict(
        etag=nsg.etag,
        id=nsg.id,
        location=nsg.location,
        name=nsg.name,
        tags=nsg.tags,
        type=nsg.type,
    )
    result['rules'] = []
    if nsg.security_rules:
        for rule in nsg.security_rules:
            result['rules'].append(create_rule_dict_from_obj(rule))

    result['default_rules'] = []
    if nsg.default_security_rules:
        for rule in nsg.default_security_rules:
            result['default_rules'].append(create_rule_dict_from_obj(rule))

    result['network_interfaces'] = []
    if nsg.network_interfaces:
        for interface in nsg.network_interfaces:
            result['network_interfaces'].append(interface.id)

    result['subnets'] = []
    if nsg.subnets:
        for subnet in nsg.subnets:
            result['subnets'].append(subnet.id)

    return result


class AzureRMSecurityGroupInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(type='str'),
            resource_group=dict(required=True, type='str'),
            tags=dict(type='list', elements='str'),
        )

        self.results = dict(
            changed=False,
        )

        self.name = None
        self.resource_group = None
        self.tags = None

        super(AzureRMSecurityGroupInfo, self).__init__(self.module_arg_spec,
                                                       supports_check_mode=True,
                                                       supports_tags=False,
                                                       facts_module=True)

    def exec_module(self, **kwargs):

        is_old_facts = self.module._name == 'azure_rm_securitygroup_facts'
        if is_old_facts:
            self.module.deprecate("The 'azure_rm_securitygroup_facts' module has been renamed to 'azure_rm_securitygroup_info'", version=(2.9, ))

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.name is not None:
            info = self.get_item()
        else:
            info = self.list_items()

        if is_old_facts:
            self.results['ansible_facts'] = {
                'azure_securitygroups': info
            }
        self.results['securitygroups'] = info

        return self.results

    def get_item(self):
        self.log('Get properties for {0}'.format(self.name))
        item = None
        result = []

        try:
            item = self.network_client.network_security_groups.get(self.resource_group, self.name)
        except ResourceNotFoundError:
            pass

        if item and self.has_tags(item.tags, self.tags):
            result = [create_network_security_group_dict(item)]

        return result

    def list_items(self):
        self.log('List all items')
        try:
            response = self.network_client.network_security_groups.list(self.resource_group)
        except Exception as exc:
            self.fail("Error listing all items - {0}".format(str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(create_network_security_group_dict(item))
        return results


def main():
    AzureRMSecurityGroupInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: networks_sm_devices_desktop_logs_info
short_description: Information module for networks _sm _devices _desktoplogs
description:
- Get all networks _sm _devices _desktoplogs.
- Return historical records of various Systems Manager network connection details for desktop devices.
version_added: '2.16.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  networkId:
    description:
    - NetworkId path parameter. Network ID.
    type: str
  deviceId:
    description:
    - DeviceId path parameter. Device ID.
    type: str
  perPage:
    description:
    - PerPage query parameter. The number of entries per page returned. Acceptable range is 3 - 1000. Default is 1000.
    type: int
  startingAfter:
    description:
    - >
      StartingAfter query parameter. A token used by the server to indicate the start of the page. Often this is a
      timestamp or an ID but it is not limited to those. This parameter should not be defined by client
      applications. The link for the first, last, prev, or next page in the HTTP Link header should define it.
    type: str
  endingBefore:
    description:
    - >
      EndingBefore query parameter. A token used by the server to indicate the end of the page. Often this is a
      timestamp or an ID but it is not limited to those. This parameter should not be defined by client
      applications. The link for the first, last, prev, or next page in the HTTP Link header should define it.
    type: str
requirements:
- meraki >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco Meraki documentation for sm getNetworkSmDeviceDesktopLogs
  description: Complete reference of the getNetworkSmDeviceDesktopLogs API.
  link: https://developer.cisco.com/meraki/api-v1/#!get-network-sm-device-desktop-logs
notes:
  - SDK Method used are
    sm.Sm.get_network_sm_device_desktop_logs,

  - Paths used are
    get /networks/{networkId}/sm/devices/{deviceId}/desktopLogs,
"""

EXAMPLES = r"""
- name: Get all networks _sm _devices _desktoplogs
  cisco.meraki.networks_sm_devices_desktop_logs_info:
    meraki_api_key: "{{meraki_api_key}}"
    meraki_base_url: "{{meraki_base_url}}"
    meraki_single_request_timeout: "{{meraki_single_request_timeout}}"
    meraki_certificate_path: "{{meraki_certificate_path}}"
    meraki_requests_proxy: "{{meraki_requests_proxy}}"
    meraki_wait_on_rate_limit: "{{meraki_wait_on_rate_limit}}"
    meraki_nginx_429_retry_wait_time: "{{meraki_nginx_429_retry_wait_time}}"
    meraki_action_batch_retry_wait_time: "{{meraki_action_batch_retry_wait_time}}"
    meraki_retry_4xx_error: "{{meraki_retry_4xx_error}}"
    meraki_retry_4xx_error_wait_time: "{{meraki_retry_4xx_error_wait_time}}"
    meraki_maximum_retries: "{{meraki_maximum_retries}}"
    meraki_output_log: "{{meraki_output_log}}"
    meraki_log_file_prefix: "{{meraki_log_file_prefix}}"
    meraki_log_path: "{{meraki_log_path}}"
    meraki_print_console: "{{meraki_print_console}}"
    meraki_suppress_logging: "{{meraki_suppress_logging}}"
    meraki_simulate: "{{meraki_simulate}}"
    meraki_be_geo_id: "{{meraki_be_geo_id}}"
    meraki_use_iterator_for_get_pages: "{{meraki_use_iterator_for_get_pages}}"
    meraki_inherit_logging_config: "{{meraki_inherit_logging_config}}"
    perPage: 0
    startingAfter: string
    endingBefore: string
    networkId: string
    deviceId: string
    total_pages: -1
    direction: next
  register: result

"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "measuredAt": "string",
        "user": "string",
        "networkDevice": "string",
        "networkDriver": "string",
        "wifiChannel": "string",
        "wifiAuth": "string",
        "wifiBssid": "string",
        "wifiSsid": "string",
        "wifiRssi": "string",
        "wifiNoise": "string",
        "dhcpServer": "string",
        "ip": "string",
        "networkMTU": "string",
        "subnet": "string",
        "gateway": "string",
        "publicIP": "string",
        "dnsServer": "string",
        "ts": "string"
      }
    ]
"""



# -*- coding: utf-8 -*-

# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
name: x509_crl_info
short_description: Retrieve information from X.509 CRLs in PEM format
version_added: 2.10.0
author:
    - Felix Fontein (@felixfontein)
description:
    - Provided a X.509 crl in PEM format, retrieve information.
    - This is a filter version of the M(community.crypto.x509_crl_info) module.
options:
    _input:
        description:
            - The content of the X.509 CRL in PEM format.
        type: string
        required: true
    list_revoked_certificates:
        description:
            - If set to V(false), the list of revoked certificates is not included in the result.
            - This is useful when retrieving information on large CRL files. Enumerating all revoked
              certificates can take some time, including serializing the result as JSON, sending it to
              the Ansible controller, and decoding it again.
        type: bool
        default: true
        version_added: 1.7.0
extends_documentation_fragment:
    - community.crypto.name_encoding
seealso:
    - module: community.crypto.x509_crl_info
    - plugin: community.crypto.to_serial
      plugin_type: filter
'''

EXAMPLES = '''
- name: Show the Organization Name of the CRL's subject
  ansible.builtin.debug:
    msg: >-
      {{
        (
          lookup('ansible.builtin.file', '/path/to/cert.pem')
          | community.crypto.x509_crl_info
        ).issuer.organizationName
      }}
'''

RETURN = '''
_value:
    description:
        - Information on the CRL.
    type: dict
    contains:
        format:
            description:
                - Whether the CRL is in PEM format (V(pem)) or in DER format (V(der)).
            returned: success
            type: str
            sample: pem
            choices:
                - pem
                - der
        issuer:
            description:
                - The CRL's issuer.
                - Note that for repeated values, only the last one will be returned.
                - See O(name_encoding) for how IDNs are handled.
            returned: success
            type: dict
            sample: {"organizationName": "Ansible", "commonName": "ca.example.com"}
        issuer_ordered:
            description: The CRL's issuer as an ordered list of tuples.
            returned: success
            type: list
            elements: list
            sample: [["organizationName", "Ansible"], ["commonName": "ca.example.com"]]
        last_update:
            description: The point in time from which this CRL can be trusted as ASN.1 TIME.
            returned: success
            type: str
            sample: '20190413202428Z'
        next_update:
            description: The point in time from which a new CRL will be issued and the client has to check for it as ASN.1 TIME.
            returned: success
            type: str
            sample: '20190413202428Z'
        digest:
            description: The signature algorithm used to sign the CRL.
            returned: success
            type: str
            sample: sha256WithRSAEncryption
        revoked_certificates:
            description: List of certificates to be revoked.
            returned: success if O(list_revoked_certificates=true)
            type: list
            elements: dict
            contains:
                serial_number:
                    description:
                        - Serial number of the certificate.
                        - This return value is an B(integer). If you need the serial numbers as a colon-separated hex string,
                          such as C(11:22:33), you need to convert it to that form with P(community.crypto.to_serial#filter).
                    type: int
                    sample: 1234
                revocation_date:
                    description: The point in time the certificate was revoked as ASN.1 TIME.
                    type: str
                    sample: '20190413202428Z'
                issuer:
                    description:
                        - The certificate's issuer.
                        - See O(name_encoding) for how IDNs are handled.
                    type: list
                    elements: str
                    sample: ["DNS:ca.example.org"]
                issuer_critical:
                    description: Whether the certificate issuer extension is critical.
                    type: bool
                    sample: false
                reason:
                    description:
                        - The value for the revocation reason extension.
                    type: str
                    sample: key_compromise
                    choices:
                        - unspecified
                        - key_compromise
                        - ca_compromise
                        - affiliation_changed
                        - superseded
                        - cessation_of_operation
                        - certificate_hold
                        - privilege_withdrawn
                        - aa_compromise
                        - remove_from_crl
                reason_critical:
                    description: Whether the revocation reason extension is critical.
                    type: bool
                    sample: false
                invalidity_date:
                    description: |
                        The point in time it was known/suspected that the private key was compromised
                        or that the certificate otherwise became invalid as ASN.1 TIME.
                    type: str
                    sample: '20190413202428Z'
                invalidity_date_critical:
                    description: Whether the invalidity date extension is critical.
                    type: bool
                    sample: false
'''

import base64
import binascii

from ansible.errors import AnsibleFilterError
from ansible.module_utils.six import string_types
from ansible.module_utils.common.text.converters import to_bytes, to_native

from ansible_collections.community.crypto.plugins.module_utils.crypto.basic import (
    OpenSSLObjectError,
)

from ansible_collections.community.crypto.plugins.module_utils.crypto.pem import (
    identify_pem_format,
)

from ansible_collections.community.crypto.plugins.module_utils.crypto.module_backends.crl_info import (
    get_crl_info,
)

from ansible_collections.community.crypto.plugins.plugin_utils.filter_module import FilterModuleMock


def x509_crl_info_filter(data, name_encoding='ignore', list_revoked_certificates=True):
    '''Extract information from X.509 PEM certificate.'''
    if not isinstance(data, string_types):
        raise AnsibleFilterError('The community.crypto.x509_crl_info input must be a text type, not %s' % type(data))
    if not isinstance(name_encoding, string_types):
        raise AnsibleFilterError('The name_encoding option must be of a text type, not %s' % type(name_encoding))
    if not isinstance(list_revoked_certificates, bool):
        raise AnsibleFilterError('The list_revoked_certificates option must be a boolean, not %s' % type(list_revoked_certificates))
    name_encoding = to_native(name_encoding)
    if name_encoding not in ('ignore', 'idna', 'unicode'):
        raise AnsibleFilterError('The name_encoding option must be one of the values "ignore", "idna", or "unicode", not "%s"' % name_encoding)

    data = to_bytes(data)
    if not identify_pem_format(data):
        try:
            data = base64.b64decode(to_native(data))
        except (binascii.Error, TypeError, ValueError, UnicodeEncodeError) as e:
            pass

    module = FilterModuleMock({'name_encoding': name_encoding})
    try:
        return get_crl_info(module, content=data, list_revoked_certificates=list_revoked_certificates)
    except OpenSSLObjectError as exc:
        raise AnsibleFilterError(to_native(exc))


class FilterModule(object):
    '''Ansible jinja2 filters'''

    def filters(self):
        return {
            'x509_crl_info': x509_crl_info_filter,
        }
